Type: Package
Package: rsynthbio
Title: Synthesize Bio API Wrapper
Version: 4.2.0
Authors@R: person(given = "Synthesize Bio",
    email = "candace@synthesize.bio",
    role = c("aut", "cre"))
Description: Access Synthesize Bio models from their API <https://app.synthesize.bio/> using
  this wrapper that provides a convenient interface to the Synthesize Bio API, allowing users
  to generate realistic gene expression data based on specified biological conditions. This package
  enables researchers to easily access AI-generated transcriptomic data for various
  modalities including bulk RNA-seq, single-cell RNA-seq, microarray data, and more.
URL: https://github.com/synthesizebio/rsynthbio
BugReports: https://github.com/synthesizebio/rsynthbio/issues
Imports: getPass, keyring, jsonlite, httr
Suggests: rmarkdown, knitr, testthat (>= 3.0.2), mockery, arrow
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
License: MIT + file LICENSE
NeedsCompilation: no
Packaged: 2026-06-11 21:28:00 UTC; alex
Author: Synthesize Bio [aut, cre]
Maintainer: Synthesize Bio <candace@synthesize.bio>
Repository: CRAN
Date/Publication: 2026-06-12 06:30:02 UTC
Built: R 4.4.3; ; 2026-06-12 07:35:25 UTC; unix

# Generated by roxygen2: do not edit by hand

export(API_BASE_URL)
export(DEFAULT_POLL_INTERVAL_SECONDS)
export(DEFAULT_POLL_TIMEOUT_SECONDS)
export(DEFAULT_TIMEOUT)
export(SELF_HOSTED_TIMEOUT)
export(clear_synthesize_token)
export(get_example_query)
export(has_synthesize_token)
export(list_models)
export(load_synthesize_token_from_keyring)
export(predict_query)
export(set_synthesize_token)
importFrom(getPass,getPass)
importFrom(httr,GET)
importFrom(httr,POST)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,http_status)
importFrom(httr,status_code)
importFrom(httr,timeout)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(utils,browseURL)
importFrom(utils,modifyList)

# <img src="https://assets.synthesize.bio/logomark.png" style="width: 30px; height: 30px;" alt="Logomark"> rsynthbio

`rsynthbio` is an R package that provides a convenient interface to the [Synthesize Bio](https://www.synthesize.bio/) API, allowing users to generate realistic gene expression data based on specified biological conditions. This package enables researchers to easily access AI-generated transcriptomic data for various modalities, including bulk RNA-seq and single-cell RNA-seq.

To generate datasets without code, use our [web platform](https://app.synthesize.bio/datasets/).

[Get started](https://docs.synthesize.bio/rsynthbio/getting-started) | [Full R SDK docs](https://docs.synthesize.bio/rsynthbio)

For function-level reference, use R's built-in help (`?predict_query`, `help(package = "rsynthbio")`) or the [CRAN reference manual](https://CRAN.R-project.org/package=rsynthbio).

For questions, suggestions, and support, email us at [support@synthesize.bio](mailto:support@synthesize.bio).

## How to install

You can install `rsynthbio` from CRAN:

```r
install.packages("rsynthbio")
```

To install the development version from GitHub:

```r
remotes::install_github("synthesizebio/rsynthbio")
```

## Authentication

Set up your API token (obtain one at [app.synthesize.bio](https://app.synthesize.bio)):

```r
library(rsynthbio)

# Securely prompt for your token (recommended)
set_synthesize_token()

# Or store in system keyring for persistence across sessions
set_synthesize_token(use_keyring = TRUE)

# Load from keyring in future sessions
load_synthesize_token_from_keyring()
```

## Quick start

```r
library(rsynthbio)

# See available models
list_models()

# Get an example query for bulk RNA-seq
query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Generate synthetic expression data
result <- predict_query(query, model_id = "gem-1-bulk")

# Access the results
metadata <- result$metadata
expression <- result$expression
```

## Available models

| Model type                 | Bulk                                  | Single-cell                          | Description                                       |
| -------------------------- | ------------------------------------- | ------------------------------------ | ------------------------------------------------- |
| **Baseline**               | `gem-1-bulk`                          | `gem-1-sc`                           | Generate expression from metadata                 |
| **Reference conditioning** | `gem-1-bulk_reference-conditioning`   | `gem-1-sc_reference-conditioning`    | Generate expression anchored to a reference sample|
| **Metadata prediction**    | `gem-1-bulk_predict-metadata`         | `gem-1-sc_predict-metadata`          | Predict metadata from expression                  |

Only baseline models are available to all users. Check programmatically with `list_models()`. Contact us at [support@synthesize.bio](mailto:support@synthesize.bio) if you have any questions.

## Self-hosted models

Partners who run Synthesize models inside their own environment can point the
same client at a self-hosted container. Self-hosted mode sends a single
synchronous request and decodes the Apache Arrow IPC stream response into the
same data frames as the hosted path (no polling, no download URL). It requires
the optional `arrow` package and does not require an API key.

Self-hosted deployment is a model deployment option available within a
Synthesize Bio partnership. To learn more or request access, contact
[partnerships@synthesize.bio](mailto:partnerships@synthesize.bio).

```r
install.packages("arrow")

# Point each model at its own container once (per-model environment variables)
Sys.setenv(
  SYNTHESIZE_SELF_HOSTED = "1",
  SYNTHESIZE_API_BASE_URL__GEM_1_BULK = "https://gem-1-bulk.internal.example",
  SYNTHESIZE_API_BASE_URL__GEM_1_SC   = "https://gem-1-sc.internal.example"
)

query <- get_example_query("gem-1-bulk", self_hosted = TRUE)$example_query
result <- predict_query(query, model_id = "gem-1-bulk", self_hosted = TRUE)
result$expression
```

You can also pass `api_base_url` explicitly per call instead of using the
environment variables. Base-URL resolution precedence is: explicit
`api_base_url` -> per-model `SYNTHESIZE_API_BASE_URL__<MODEL>` -> global
`SYNTHESIZE_API_BASE_URL` -> production default. If the container is started
with authentication enabled, set `SYNTHESIZE_API_KEY` and the client will send
it as a bearer token. See the [Self-Hosted Models](https://docs.synthesize.bio/rsynthbio/self-hosted)
vignette for details.

## Documentation

For detailed usage and guides, see the [R SDK section of the Synthesize Bio docs](https://docs.synthesize.bio/rsynthbio):

- [Getting started](https://docs.synthesize.bio/rsynthbio/getting-started) — Installation, authentication, and overview
- [Available metadata](https://docs.synthesize.bio/rsynthbio/available-metadata) — Metadata fields you can query
- [Baseline models](https://docs.synthesize.bio/rsynthbio/models/baseline) — Generate expression from metadata
- [Reference conditioning](https://docs.synthesize.bio/rsynthbio/models/reference-conditioning) — Condition on real expression data
- [Metadata prediction](https://docs.synthesize.bio/rsynthbio/models/metadata-prediction) — Infer metadata from expression
- [Self-hosted models](https://docs.synthesize.bio/rsynthbio/self-hosted) — Run models in your own environment via Arrow streaming
- [Function reference](https://docs.synthesize.bio/rsynthbio/reference) — All exported functions

The legacy pkgdown site at `synthesizebio.github.io/rsynthbio` redirects to the
canonical docs above; see `docs-redirect/` and the `deploy-docs-redirect`
workflow for how those redirects are served.

## Authoring the Mintlify docs

The shared docs site at
[docs.synthesize.bio](https://docs.synthesize.bio) aggregates the R SDK
section directly from this repo into `/rsynthbio/...` — there is no
separate docs repo to update.

Authoring sources:

- Long-form pages: `vignettes/*.Rmd` (also shipped as R vignettes for offline
  `vignette(package = "rsynthbio")` browsing)
- Function reference: `man/*.Rd` (generated from `R/*.R` roxygen comments)
- Generation script: `scripts/generate_mintlify_docs.py`

Regenerate the Mintlify output in `docs-external/` after changing vignettes,
roxygen docs, or the package surface:

```bash
python3 scripts/generate_mintlify_docs.py
```

The generated pages are committed in `docs-external/` so changes are
reviewable in PRs.

## Rate limits

Free usage of Synthesize Bio is limited. If you exceed this limit, you'll receive an error message. To generate more samples, please contact us at [support@synthesize.bio](mailto:support@synthesize.bio).

# <img src="https://assets.synthesize.bio/logomark.png" style="width: 30px; height: 30px;" alt="Logomark"> rsynthbio

`rsynthbio` is an R package that provides a convenient interface to the [Synthesize Bio](https://www.synthesize.bio/) API, allowing users to generate realistic gene expression data based on specified biological conditions. This package enables researchers to easily access AI-generated transcriptomic data for various modalities, including bulk RNA-seq and single-cell RNA-seq.

To generate datasets without code, use our [web platform](https://app.synthesize.bio/datasets/).

[See the full documentation here](https://synthesizebio.github.io/rsynthbio/).

For questions, suggestions, and support, email us at [support@synthesize.bio](mailto:support@synthesize.bio).

## How to Install

You can install `rsynthbio` from CRAN:

```r
install.packages("rsynthbio")
```

To install the development version from GitHub:

```r
remotes::install_github("synthesizebio/rsynthbio")
```

## Authentication

Set up your API token (obtain one at [app.synthesize.bio](https://app.synthesize.bio)):

```r
library(rsynthbio)

# Securely prompt for your token (recommended)
set_synthesize_token()

# Or store in system keyring for persistence across sessions
set_synthesize_token(use_keyring = TRUE)

# Load from keyring in future sessions
load_synthesize_token_from_keyring()
```

## Quick Start

```r
library(rsynthbio)

# See available models
list_models()

# Get an example query for bulk RNA-seq
query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Generate synthetic expression data
result <- predict_query(query, model_id = "gem-1-bulk")

# Access the results
metadata <- result$metadata
expression <- result$expression
```

## Available Models

| Model Type | Bulk | Single-Cell | Description |
|------------|------|-------------|-------------|
| **Baseline** | `gem-1-bulk` | `gem-1-sc` | Generate expression from metadata |
| **Reference Conditioning** | `gem-1-bulk_reference-conditioning` | `gem-1-sc_reference-conditioning` | Generate expression anchored to a reference sample |
| **Metadata Prediction** | `gem-1-bulk_predict-metadata` | `gem-1-sc_predict-metadata` | Predict metadata from expression |

Only baseline models are available to all users. You can check which models are available programmatically, use `list_models()`. Contact us at support@synthesize.bio if you have any questions.

## Documentation

For detailed usage instructions, see the vignettes:

- [Getting Started](https://synthesizebio.github.io/rsynthbio/articles/getting-started.html) — Installation, authentication, and overview
- [Baseline Models](https://synthesizebio.github.io/rsynthbio/articles/baseline.html) — Generate expression from metadata
- [Reference Conditioning](https://synthesizebio.github.io/rsynthbio/articles/reference-conditioning.html) — Condition on real expression data
- [Metadata Prediction](https://synthesizebio.github.io/rsynthbio/articles/metadata-prediction.html) — Infer metadata from expression

## Rate Limits

Free usage of Synthesize Bio is limited. If you exceed this limit, you will receive an error message. To generate more samples, please contact us at [support@synthesize.bio](mailto:support@synthesize.bio).

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)

## -----------------------------------------------------------------------------
# library(rsynthbio)

## ----query-example, eval=FALSE------------------------------------------------
# # Get the example query structure
# example_query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query
# 
# # Inspect the query structure
# str(example_query)

## ----predict-metadata, eval=FALSE---------------------------------------------
# # Start with example query structure
# query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query
# 
# # Replace with your actual expression counts
# # Each input should be a list with a counts vector
# query$inputs <- list(
#   list(counts = sample1_counts),
#   list(counts = sample2_counts),
#   list(counts = sample3_counts)
# )
# 
# # Optional: set seed for reproducibility
# query$seed <- 42
# 
# # Submit the query
# result <- predict_query(query, model_id = "gem-1-bulk_predict-metadata")

## ----single-sample, eval=FALSE------------------------------------------------
# query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query
# 
# # Single sample
# query$inputs <- list(
#   list(counts = my_sample_counts)
# )
# 
# result <- predict_query(query, model_id = "gem-1-bulk_predict-metadata")
# 
# # Access the predictions
# print(result$outputs$metadata)

## ----inputs-example, eval=FALSE-----------------------------------------------
# query$inputs <- list(
#   list(counts = c(0, 12, 5, 0, 33, 7, ...)), # Sample 1
#   list(counts = c(3, 0, 0, 7, 1, 0, ...)) # Sample 2
# )

## ----seed-example, eval=FALSE-------------------------------------------------
# query$seed <- 123

## ----metadata-results, eval=FALSE---------------------------------------------
# # View predicted metadata
# head(result$outputs$metadata)
# 
# # Access specific predictions
# result$outputs$metadata$cell_type_ontology_id
# result$outputs$metadata$tissue_ontology_id
# result$outputs$metadata$disease_ontology_id

## ----classifier-probs, eval=FALSE---------------------------------------------
# # If probabilities are included in the output
# # Access cell type probabilities for first sample
# # The exact structure depends on the API response format
# 
# # Example: viewing top predicted cell types
# cell_type_probs <- result$outputs$classifier_probs$cell_type[[1]]
# head(sort(cell_type_probs, decreasing = TRUE))

## ----latents, eval=FALSE------------------------------------------------------
# # Access latent representations (if returned)
# biological_latents <- result$outputs$latents$biological
# technical_latents <- result$outputs$latents$technical

## ----annotation, eval=FALSE---------------------------------------------------
# # Load your unlabeled samples
# unlabeled_counts <- read.csv("unlabeled_samples.csv", row.names = 1)
# 
# # Create query
# query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query
# query$inputs <- lapply(1:ncol(unlabeled_counts), function(i) {
#   list(counts = unlabeled_counts[, i])
# })
# 
# # Predict metadata
# result <- predict_query(query, model_id = "gem-1-bulk_predict-metadata")
# 
# # Combine with sample IDs
# annotations <- result$outputs$metadata
# annotations$sample_id <- colnames(unlabeled_counts)

## ----qc, eval=FALSE-----------------------------------------------------------
# # Compare predicted vs. provided labels
# provided_labels <- c("UBERON:0002107", "UBERON:0002107", "UBERON:0000955", "UBERON:0000955")
# predicted_labels <- result$outputs$metadata$tissue_ontology_id
# 
# # Identify potential mismatches
# mismatches <- which(provided_labels != predicted_labels)
# if (length(mismatches) > 0) {
#   message("Potential mislabeled samples: ", paste(mismatches, collapse = ", "))
# }

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since these calls require a running container
)

## ----install, eval=FALSE------------------------------------------------------
# install.packages("rsynthbio")
# install.packages("arrow") # required for self-hosted mode

## ----enable, eval=FALSE-------------------------------------------------------
# library(rsynthbio)
# 
# Sys.setenv(SYNTHESIZE_SELF_HOSTED = "1")

## ----per-model, eval=FALSE----------------------------------------------------
# Sys.setenv(
#   SYNTHESIZE_API_BASE_URL__GEM_1_BULK = "https://gem-1-bulk.internal.example",
#   SYNTHESIZE_API_BASE_URL__GEM_1_SC   = "https://gem-1-sc.internal.example"
# )
# 
# query <- get_example_query("gem-1-bulk", self_hosted = TRUE)$example_query
# result <- predict_query(query, model_id = "gem-1-bulk", self_hosted = TRUE)
# 
# expression <- result$expression
# metadata <- result$metadata

## ----explicit-url, eval=FALSE-------------------------------------------------
# result <- predict_query(
#   query,
#   model_id = "gem-1-bulk",
#   api_base_url = "https://gem-1-bulk.internal.example",
#   self_hosted = TRUE
# )

## ----auth, eval=FALSE---------------------------------------------------------
# Sys.setenv(SYNTHESIZE_API_KEY = "your-container-api-key")

## ----raw, eval=FALSE----------------------------------------------------------
# raw <- predict_query(
#   query,
#   model_id = "gem-1-bulk",
#   self_hosted = TRUE,
#   raw_response = TRUE
# )
# 
# raw$table # an arrow::Table
# raw$model_version

## ----session-info-------------------------------------------------------------
# sessionInfo()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)

## -----------------------------------------------------------------------------
# library(rsynthbio)

## ----query-example, eval=FALSE------------------------------------------------
# # Get the example query structure
# example_query <- get_example_query(model_id = "gem-1-bulk_reference-conditioning")$example_query
# 
# # Inspect the query structure
# str(example_query)

## ----drug-treatment, eval=FALSE-----------------------------------------------
# # Start with example query structure
# query <- get_example_query(model_id = "gem-1-bulk_reference-conditioning")$example_query
# 
# # Replace with your actual reference counts
# # The counts vector must match the model's expected gene order and length
# query$inputs[[1]]$counts <- your_reference_counts
# 
# # Specify the perturbation
# query$inputs[[1]]$metadata <- list(
#   perturbation_ontology_id = "CHEMBL25", # Aspirin (ChEMBL ID)
#   perturbation_type = "compound",
#   perturbation_time = "24 hours",
#   perturbation_dose = "10 um"
# )
# 
# query$inputs[[1]]$num_samples <- 3
# 
# # Set the sampling strategy
# query$sampling_strategy <- "mean estimation"
# 
# # Submit the query
# result <- predict_query(query, model_id = "gem-1-bulk_reference-conditioning")

## ----crispr-example, eval=FALSE-----------------------------------------------
# query <- get_example_query(model_id = "gem-1-bulk_reference-conditioning")$example_query
# 
# # Your reference sample counts
# query$inputs[[1]]$counts <- control_sample_counts
# 
# # CRISPR knockout of TP53
# query$inputs[[1]]$metadata <- list(
#   perturbation_ontology_id = "ENSG00000141510", # TP53 Ensembl ID
#   perturbation_type = "crispr"
# )
# 
# query$inputs[[1]]$num_samples <- 5
# 
# result <- predict_query(query, model_id = "gem-1-bulk_reference-conditioning")

## ----mode-example, eval=FALSE-------------------------------------------------
# query$sampling_strategy <- "mean estimation"

## ----fixed-total-count, eval=FALSE--------------------------------------------
# # Preserve reference library size (default)
# query$fixed_total_count <- FALSE
# 
# # Or force a specific library size
# query$fixed_total_count <- TRUE
# query$total_count <- 10000000

## ----deterministic-example, eval=FALSE----------------------------------------
# query$deterministic_latents <- TRUE

## ----seed-example, eval=FALSE-------------------------------------------------
# query$seed <- 42

## ----results, eval=FALSE------------------------------------------------------
# # Access metadata and expression matrices
# metadata <- result$metadata
# expression <- result$expression
# 
# # Compare to your reference
# dim(expression)
# head(metadata)

## ----de-analysis, eval=FALSE--------------------------------------------------
# # Your reference (input) counts
# reference_cpm <- your_reference_counts / sum(your_reference_counts) * 1e6
# 
# # Generated (perturbed) counts
# generated_cpm <- expression[1, ] / sum(expression[1, ]) * 1e6
# 
# # Log fold change
# log2fc <- log2(generated_cpm + 1) - log2(reference_cpm + 1)
# 
# # Identify top changed genes
# head(sort(log2fc, decreasing = TRUE), 20)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)

## ----installation, eval=FALSE-------------------------------------------------
# install.packages("rsynthbio")

## ----github-installation, eval=FALSE------------------------------------------
# if (!("remotes" %in% installed.packages())) {
#   install.packages("remotes")
# }
# remotes::install_github("synthesizebio/rsynthbio")

## -----------------------------------------------------------------------------
# library(rsynthbio)

## ----auth-secure, eval=FALSE--------------------------------------------------
# # Securely prompt for and store your API token
# # The token will not be visible in the console
# set_synthesize_token()
# 
# # You can also store the token in your system keyring for persistence
# # across R sessions (requires the 'keyring' package)
# set_synthesize_token(use_keyring = TRUE)

## ----eval=FALSE---------------------------------------------------------------
# # In future sessions, load the stored token
# load_synthesize_token_from_keyring()
# 
# # Check if a token is already set
# has_synthesize_token()

## ----eval=FALSE---------------------------------------------------------------
# set_synthesize_token(token = "your-token-here")

## ----models-------------------------------------------------------------------
# # Check available models
# list_models()

## ----query, eval=FALSE--------------------------------------------------------
# # Get an example query structure
# query <- get_example_query(model_id = "gem-1-bulk")$example_query
# 
# # Submit the query and get results
# result <- predict_query(query, model_id = "gem-1-bulk")
# 
# # Access the results
# metadata <- result$metadata
# expression <- result$expression

## ----session-info-------------------------------------------------------------
# sessionInfo()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)

## -----------------------------------------------------------------------------
# library(rsynthbio)

## ----query-example------------------------------------------------------------
# # Get the example query structure for a specific model
# example_query <- get_example_query(model_id = "gem-1-bulk")$example_query
# 
# # Inspect the query structure
# str(example_query)

## ----predict, eval=FALSE------------------------------------------------------
# # Create a query for the bulk model
# query <- get_example_query(model_id = "gem-1-bulk")$example_query
# 
# # Submit and get results
# result <- predict_query(query, model_id = "gem-1-bulk")

## ----sc-example, eval=FALSE---------------------------------------------------
# # Create a query for the single-cell model
# sc_query <- get_example_query(model_id = "gem-1-sc")$example_query
# 
# # Submit and get results
# sc_result <- predict_query(sc_query, model_id = "gem-1-sc")

## ----mode-examples, eval=FALSE------------------------------------------------
# # Bulk query with sample generation
# bulk_query <- get_example_query(model_id = "gem-1-bulk")$example_query
# bulk_query$sampling_strategy <- "sample generation"
# 
# # Bulk query with mean estimation
# bulk_query_mean <- get_example_query(model_id = "gem-1-bulk")$example_query
# bulk_query_mean$sampling_strategy <- "mean estimation"
# 
# # Single-cell query (must use mean estimation)
# sc_query <- get_example_query(model_id = "gem-1-sc")$example_query
# sc_query$sampling_strategy <- "mean estimation" # Required for single-cell

## ----total-count-example, eval=FALSE------------------------------------------
# # Create a query and add custom total_count
# query <- get_example_query(model_id = "gem-1-bulk")$example_query
# query$total_count <- 5000000

## ----deterministic-example, eval=FALSE----------------------------------------
# # Create a query and enable deterministic latents
# query <- get_example_query(model_id = "gem-1-bulk")$example_query
# query$deterministic_latents <- TRUE

## ----seed-example, eval=FALSE-------------------------------------------------
# # Create a query with a specific seed
# query <- get_example_query(model_id = "gem-1-bulk")$example_query
# query$seed <- 42

## ----combined-params, eval=FALSE----------------------------------------------
# # Create a query and add multiple parameters
# query <- get_example_query(model_id = "gem-1-bulk")$example_query
# query$total_count <- 8000000
# query$deterministic_latents <- TRUE
# query$sampling_strategy <- "mean estimation"
# 
# results <- predict_query(query, model_id = "gem-1-bulk")

## ----modify-query, eval=FALSE-------------------------------------------------
# # Get a base query
# query <- get_example_query(model_id = "gem-1-bulk")$example_query
# 
# # Adjust number of samples for the first input
# query$inputs[[1]]$num_samples <- 10
# 
# # Add a new condition
# query$inputs[[3]] <- list(
#   metadata = list(
#     sex = "male",
#     sample_type = "primary tissue",
#     tissue_ontology_id = "UBERON:0002371"
#   ),
#   num_samples = 5
# )

## ----analyze, eval=FALSE------------------------------------------------------
# # Access metadata and expression matrices
# metadata <- result$metadata
# expression <- result$expression
# 
# # Check dimensions
# dim(expression)
# 
# # View metadata sample
# head(metadata)

## ----large-data, eval=FALSE---------------------------------------------------
# # Save results to RDS file
# saveRDS(result, "synthesize_results.rds")
# 
# # Load previously saved results
# result <- readRDS("synthesize_results.rds")
# 
# # Export as CSV
# write.csv(result$expression, "expression_matrix.csv")
# write.csv(result$metadata, "sample_metadata.csv")

#' @title List Available Models
#' @description Returns a list of all models available in the Synthesize Bio API.
#' Each model has a unique ID that can be used with predict_query() and get_example_query().
#'
#' @param api_base_url The base URL for the API server. When NULL (default), it
#'        is resolved from the `SYNTHESIZE_API_BASE_URL` environment variable,
#'        falling back to the production default (API_BASE_URL). Point this at a
#'        self-hosted model container to list its models.
#' @param self_hosted Logical; when TRUE, the request targets a self-hosted
#'        container and does not require an API key (one is only sent if set).
#'        When NULL (default), it is resolved from the `SYNTHESIZE_SELF_HOSTED`
#'        environment variable (truthy for 1/true/yes/on), defaulting to FALSE.
#' @return A list or data frame containing available models with their IDs and metadata.
#' @examples
#' \dontrun{
#' # Get all available models
#' models <- list_models()
#' print(models)
#'
#' # List models from a self-hosted container (no API key required)
#' models <- list_models(
#'   api_base_url = "https://gem-1-bulk.internal.partner.example",
#'   self_hosted = TRUE
#' )
#' }
#' @importFrom httr GET add_headers content status_code timeout
#' @importFrom jsonlite fromJSON
#' @export
list_models <- function(api_base_url = NULL, self_hosted = NULL) {
  api_base_url <- resolve_api_base_url(api_base_url)
  self_hosted <- resolve_self_hosted(self_hosted)
  url <- paste0(api_base_url, "/api/models")
  return(make_api_request(url, "List models", self_hosted = self_hosted))
}

#' @title Get Example Query for Model
#' @description Retrieves an example query structure for a specific model.
#' This provides a template that can be modified for your specific needs.
#'
#' @param model_id Character string specifying the model ID (e.g., "gem-1-bulk", "gem-1-sc").
#' @param api_base_url The base URL for the API server. When NULL (default), it
#'        is resolved from the `SYNTHESIZE_API_BASE_URL` environment variable,
#'        falling back to the production default (API_BASE_URL). Point this at a
#'        self-hosted model container to fetch its example query.
#' @param self_hosted Logical; when TRUE, the request targets a self-hosted
#'        container and does not require an API key (one is only sent if set).
#'        When NULL (default), it is resolved from the `SYNTHESIZE_SELF_HOSTED`
#'        environment variable (truthy for 1/true/yes/on), defaulting to FALSE.
#' @return A list representing a valid query structure for the specified model.
#' @examples
#' \dontrun{
#' # Get example query for bulk RNA-seq model
#' query <- get_example_query(model_id = "gem-1-bulk")$example_query
#'
#' # Get example query for single-cell model
#' query_sc <- get_example_query(model_id = "gem-1-sc")$example_query
#'
#' # Modify the query structure
#' query$inputs[[1]]$num_samples <- 10
#'
#' # Fetch from a self-hosted container (no API key required)
#' query <- get_example_query(
#'   model_id = "gem-1-bulk",
#'   api_base_url = "https://gem-1-bulk.internal.partner.example",
#'   self_hosted = TRUE
#' )$example_query
#' }
#' @importFrom httr GET add_headers content status_code timeout
#' @importFrom jsonlite fromJSON
#' @export
get_example_query <- function(model_id, api_base_url = NULL, self_hosted = NULL) {
  api_base_url <- resolve_api_base_url(api_base_url, model_id = model_id)
  self_hosted <- resolve_self_hosted(self_hosted)
  url <- paste0(api_base_url, "/api/models/", model_id, "/example-query")
  return(make_api_request(url, "Get example query", self_hosted = self_hosted))
}

#' @title Start Model Query
#' @description Internal function to start an async model query
#' @param api_base_url The base URL for the API
#' @param model_id The model ID for the specific model
#' @param query The query list
#' @return The model query ID
#' @importFrom httr POST add_headers content status_code timeout
#' @importFrom jsonlite toJSON fromJSON
#' @noRd
start_model_query <- function(api_base_url, model_id, query) {
  url <- paste0(api_base_url, "/api/models/", model_id, "/predict")
  query_json <- toJSON(query, auto_unbox = TRUE)

  response <- tryCatch(
    {
      POST(
        url = url,
        add_headers(
          Accept = "application/json",
          Authorization = paste("Bearer", Sys.getenv("SYNTHESIZE_API_KEY")),
          `Content-Type` = "application/json"
        ),
        body = query_json,
        encode = "json",
        timeout(DEFAULT_TIMEOUT)
      )
    },
    error = function(e) {
      stop(paste0("Predict request failed due to a network issue: ", e$message))
    }
  )

  if (status_code(response) >= 400) {
    stop(paste0(
      "Predict request failed with status ",
      status_code(response), ": ", content(response, "text")
    ))
  }

  parsed_content <- tryCatch(
    {
      fromJSON(content(response, "text"), simplifyDataFrame = TRUE)
    },
    error = function(e) {
      stop(paste0("Failed to decode JSON from predict response: ", e$message))
    }
  )

  if (!is.list(parsed_content) || is.null(parsed_content$modelQueryId)) {
    stop(paste0(
      "Unexpected response from predict endpoint: ",
      paste(names(parsed_content), collapse = ", ")
    ))
  }

  return(as.character(parsed_content$modelQueryId))
}

#' @title Poll Model Query
#' @description Internal function to poll the status endpoint until ready/failed or timeout
#' @param api_base_url The base URL for the API
#' @param model_query_id The model query ID to poll
#' @param poll_interval Seconds between polling attempts
#' @param timeout_seconds Maximum total seconds to wait
#' @return A list with status and payload
#' @importFrom httr GET add_headers content status_code timeout
#' @importFrom jsonlite fromJSON
#' @noRd
poll_model_query <- function(api_base_url, model_query_id, poll_interval, timeout_seconds) {
  start_time <- Sys.time()
  status_url <- paste0(api_base_url, "/api/model-queries/", model_query_id, "/status")
  last_payload <- list()

  while (TRUE) {
    response <- tryCatch(
      {
        GET(
          url = status_url,
          add_headers(
            Accept = "application/json",
            Authorization = paste("Bearer", Sys.getenv("SYNTHESIZE_API_KEY"))
          ),
          timeout(DEFAULT_TIMEOUT)
        )
      },
      error = function(e) {
        stop(paste0("Status request failed due to a network issue: ", e$message))
      }
    )

    if (status_code(response) >= 400) {
      stop(paste0(
        "Status request failed with status ",
        status_code(response), ": ", content(response, "text")
      ))
    }

    payload <- tryCatch(
      {
        fromJSON(content(response, "text"), simplifyDataFrame = TRUE)
      },
      error = function(e) {
        stop(paste0("Failed to decode JSON from status response: ", e$message))
      }
    )

    if (!is.list(payload) || is.null(payload$status)) {
      stop(paste0(
        "Unexpected status response: ",
        paste(names(payload), collapse = ", ")
      ))
    }

    status <- as.character(payload$status)
    last_payload <- payload

    if (status %in% c("ready", "failed")) {
      return(list(status = status, payload = payload))
    }

    elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
    if (elapsed > timeout_seconds) {
      return(list(status = status, payload = last_payload))
    }

    Sys.sleep(max(1, as.integer(poll_interval)))
  }
}

#' @title Get JSON from URL
#' @description Internal function to fetch JSON from a URL
#' @param url The URL to fetch from
#' @return The parsed JSON content
#' @importFrom httr GET content status_code timeout
#' @importFrom jsonlite fromJSON
#' @noRd
get_json <- function(url) {
  response <- tryCatch(
    {
      GET(url, timeout(DEFAULT_TIMEOUT))
    },
    error = function(e) {
      stop(paste0("Failed to fetch download URL due to a network issue: ", e$message))
    }
  )

  if (status_code(response) >= 400) {
    stop(paste0(
      "Download URL fetch failed with status ",
      status_code(response), ": ", content(response, "text")
    ))
  }

  parsed_content <- tryCatch(
    {
      fromJSON(
        content(response, "text", encoding = "UTF-8"),
        simplifyDataFrame = TRUE
      )
    },
    error = function(e) {
      stop(paste0("Failed to decode JSON from download URL response: ", e$message))
    }
  )

  return(parsed_content)
}

#' @title Build Headers for a JSON API Request (Internal)
#' @description Always requests JSON. Attaches an `Authorization` header when an
#' API key is configured. In `self_hosted` mode the header is omitted when no key
#' is set, so the client works against containers with authentication disabled.
#' @param self_hosted Logical; when TRUE, allow requests without a key.
#' @return An httr `request` object produced by [httr::add_headers()].
#' @importFrom httr add_headers
#' @noRd
build_auth_headers <- function(self_hosted = FALSE) {
  api_key <- Sys.getenv("SYNTHESIZE_API_KEY")
  if (self_hosted && !nzchar(api_key)) {
    return(add_headers(Accept = "application/json"))
  }
  add_headers(
    Accept = "application/json",
    Authorization = paste("Bearer", api_key)
  )
}

#' @title Make Authenticated API Request
#' @description Internal helper to make authenticated GET requests
#' @param url The URL to fetch
#' @param context_msg String describing the request context for error messages
#' @param self_hosted Logical; when TRUE, do not require an API key (used for
#'        self-hosted containers running with authentication disabled).
#' @return The parsed JSON content
#' @importFrom httr GET add_headers content status_code timeout
#' @importFrom jsonlite fromJSON
#' @noRd
make_api_request <- function(url, context_msg, self_hosted = FALSE) {
  if (!self_hosted && !has_synthesize_token()) {
    stop("Please set your API key for Synthesize Bio using set_synthesize_token()")
  }

  response <- tryCatch(
    {
      GET(
        url = url,
        build_auth_headers(self_hosted = self_hosted),
        timeout(DEFAULT_TIMEOUT)
      )
    },
    error = function(e) {
      stop(paste0(context_msg, " request failed due to a network issue: ", e$message))
    }
  )

  if (status_code(response) >= 400) {
    stop(paste0(
      context_msg, " request failed with status ",
      status_code(response), ": ", content(response, "text")
    ))
  }

  parsed_content <- tryCatch(
    {
      fromJSON(content(response, "text"), simplifyDataFrame = TRUE)
    },
    error = function(e) {
      stop(paste0("Failed to decode JSON from ", tolower(context_msg), " response: ", e$message))
    }
  )

  return(parsed_content)
}


#' @title Predict Gene Expression
#' @description Sends a query to the Synthesize Bio API for prediction
#' and retrieves gene expression samples. This function sends the query
#' to the API and processes the response into usable data frames.
#'
#' @param query A list representing the query data to send to the API.
#'        Use `get_example_query()` to generate an example. The query supports additional
#'        optional fields:
#'        \itemize{
#'          \item `total_count` (integer): Library size used when converting predicted log CPM
#'                back to raw counts. Higher values scale counts up proportionally.
#'          \item `deterministic_latents` (logical): If TRUE, the model uses the mean of each
#'                latent distribution instead of sampling, producing deterministic outputs for
#'                the same inputs. Useful for reproducibility.
#'          \item `seed` (integer): Random seed for reproducibility.
#'        }
#' @param model_id Character string specifying the model ID (e.g., "gem-1-bulk", "gem-1-sc").
#'        Use `list_models()` to see available models.
#' @param api_base_url The base URL for the API server. When NULL (default), it
#'        is resolved in order from the per-model environment variable
#'        `SYNTHESIZE_API_BASE_URL__<MODEL>` (e.g.
#'        `SYNTHESIZE_API_BASE_URL__GEM_1_BULK`), then the global
#'        `SYNTHESIZE_API_BASE_URL`, then the production default (API_BASE_URL).
#'        The per-model variable lets you point each self-hosted model at its own
#'        container once and omit `api_base_url` on every call.
#' @param poll_interval_seconds Seconds between polling attempts of the status endpoint.
#'        Default is DEFAULT_POLL_INTERVAL_SECONDS (2).
#' @param poll_timeout_seconds Maximum total seconds to wait before timing out.
#'        Default is DEFAULT_POLL_TIMEOUT_SECONDS (900 = 15 minutes).
#' @param return_download_url Logical, if TRUE, returns a list containing the signed
#'        download URL instead of parsing into data frames. Default is FALSE.
#' @param raw_response Logical, if TRUE, returns the raw (unformatted) response
#'        from the API without applying any output transformers. For the
#'        production path this is the parsed JSON; for `self_hosted = TRUE` it is
#'        the parsed Arrow `Table` together with its schema metadata. Default is FALSE.
#' @param self_hosted Logical, if TRUE, sends a single synchronous request to a
#'        self-hosted model container that returns predictions as an Apache Arrow
#'        IPC stream (no polling, no download URL). Requires the optional `arrow`
#'        package and an `api_base_url` pointing at the container. Unlike the
#'        production path, no API key is required (one is only sent if configured).
#'        When NULL (default), it is resolved from the `SYNTHESIZE_SELF_HOSTED`
#'        environment variable (truthy for 1/true/yes/on), defaulting to FALSE.
#' @param ... Additional parameters to include in the query body. These are passed
#'        directly to the API and validated server-side.
#' @return A list. For the production path, if `return_download_url` is `FALSE`
#'         (default) the list contains `metadata` and `expression` data frames;
#'         if `TRUE` it contains `download_url` and empty data frames. For
#'         `self_hosted = TRUE`, the list contains the transformed data frames
#'         (`metadata`, `expression`, and `latents`; plus `classifier_probs` for
#'         metadata-prediction models) with `model_version` and `request_type`
#'         attached as attributes.
#' @importFrom httr POST GET add_headers content http_status status_code timeout
#' @importFrom jsonlite toJSON fromJSON
#' @importFrom utils modifyList
#' @examples
#' # Set your API key (in practice, use a more secure method)
#' \dontrun{
#'
#' # To start using rsynthbio, first you need to have an account with synthesize.bio.
#' # Go here to create one: https://app.synthesize.bio/
#'
#' set_synthesize_token()
#'
#' # Get available models
#' models <- list_models()
#'
#' # Create a query for a specific model
#' query <- get_example_query(model_id = "gem-1-bulk")$example_query
#'
#' # Request raw counts
#' result <- predict_query(query, model_id = "gem-1-bulk")
#'
#' # Access the results
#' metadata <- result$metadata
#' expression <- result$expression
#'
#' # Explore the top expressed genes in the first sample
#' head(sort(expression[1, ], decreasing = TRUE))
#'
#' # Use deterministic latents for reproducible results
#' query$deterministic_latents <- TRUE
#' result_det <- predict_query(query, model_id = "gem-1-bulk")
#'
#' # Specify a custom total count (library size)
#' query$total_count <- 5000000
#' result_custom <- predict_query(query, model_id = "gem-1-bulk")
#'
#' # Self-hosted container returning a synchronous Apache Arrow IPC stream
#' result_sh <- predict_query(
#'   query,
#'   model_id = "gem-1-bulk",
#'   api_base_url = "https://gem-1-bulk.internal.partner.example",
#'   self_hosted = TRUE
#' )
#' }
#' @export
predict_query <- function(query,
                          model_id,
                          api_base_url = NULL,
                          poll_interval_seconds = DEFAULT_POLL_INTERVAL_SECONDS,
                          poll_timeout_seconds = DEFAULT_POLL_TIMEOUT_SECONDS,
                          return_download_url = FALSE,
                          raw_response = FALSE,
                          self_hosted = NULL,
                          ...) {
  api_base_url <- resolve_api_base_url(api_base_url, model_id = model_id)
  self_hosted <- resolve_self_hosted(self_hosted)

  # Validate base URL
  if (!grepl("^https?://", api_base_url)) {
    stop(
      paste0(
        "Invalid api_base_url: ",
        api_base_url,
        ". Must start with http:// or https://"
      )
    )
  }

  # Add source field for reporting
  query$source <- "rsynthbio"

  # Merge any additional ... args into the query (validated server-side)
  extra_args <- list(...)
  if (length(extra_args) > 0) {
    query <- modifyList(query, extra_args)
  }

  # Self-hosted containers return predictions synchronously as an Arrow IPC
  # stream and do not require an API key.
  if (isTRUE(self_hosted)) {
    return(predict_query_self_hosted(
      query = query,
      model_id = model_id,
      api_base_url = api_base_url,
      raw_response = raw_response
    ))
  }

  if (!has_synthesize_token()) {
    stop("Please set your API key for Synthesize Bio using set_synthesize_token()")
  }

  # Start async query
  model_query_id <- start_model_query(
    api_base_url = api_base_url,
    model_id = model_id,
    query = query
  )

  # Poll for completion
  poll_result <- poll_model_query(
    api_base_url = api_base_url,
    model_query_id = model_query_id,
    poll_interval = poll_interval_seconds,
    timeout_seconds = poll_timeout_seconds
  )

  status <- poll_result$status
  payload <- poll_result$payload

  if (status == "failed") {
    # payload contains message if available
    err_msg <- if (is.list(payload) && !is.null(payload$message)) {
      payload$message
    } else {
      NULL
    }
    stop(paste0(
      "Model query failed. ",
      if (!is.null(err_msg)) err_msg else "No error message provided."
    ))
  }

  if (status != "ready") {
    stop(paste0(
      "Model query did not complete in time (status=", status, "). ",
      "Consider increasing poll_timeout_seconds."
    ))
  }

  # When ready, payload should contain a signed downloadUrl to the final JSON
  download_url <- if (is.list(payload) && !is.null(payload$downloadUrl)) {
    payload$downloadUrl
  } else {
    NULL
  }
  if (is.null(download_url)) {
    stop("Response missing downloadUrl when status=ready")
  }

  if (return_download_url) {
    # Caller wants the URL only; return in a structured payload
    return(list(
      metadata = data.frame(),
      expression = data.frame(),
      download_url = download_url
    ))
  }

  # Fetch the final results JSON and transform to data frames
  final_json <- get_json(download_url)

  if (isTRUE(raw_response)) {
    return(final_json)
  }

  transformer <- get_output_transformer(model_id)
  if (is.null(transformer)) {
    stop(paste0(
      "No output formatter registered for model_id '", model_id, "'. ",
      "To receive raw (unformatted) JSON, pass raw_response = TRUE ",
      "to predict_query()."
    ))
  }

  return(transformer(final_json))
}

#' @title API Base URL
#' @description Base URL for the Synthesize Bio API
#' @export
API_BASE_URL <- Sys.getenv("SYNTHESIZE_API_BASE_URL", unset = "https://app.synthesize.bio")

#' @title Default Timeout
#' @description Default timeout (seconds) for outbound HTTP requests
#' @export
DEFAULT_TIMEOUT <- 300

#' @title Self-Hosted Timeout
#' @description Timeout (seconds) for synchronous self-hosted container
#' predictions. These run on the partner's GPU box and can take minutes for
#' large sample counts, so they use a longer timeout than hosted control-plane
#' calls.
#' @export
SELF_HOSTED_TIMEOUT <- 600

#' @title Default Poll Interval
#' @description Default polling interval (seconds) for async model queries
#' @export
DEFAULT_POLL_INTERVAL_SECONDS <- 2

#' @title Default Poll Timeout
#' @description Default maximum timeout (seconds) for async model queries
#' @export
DEFAULT_POLL_TIMEOUT_SECONDS <- 15 * 60

#' @title Interpret an Environment Variable as a Boolean Flag (Internal)
#' @description Returns TRUE for the values 1/true/yes/on (case-insensitive,
#' surrounding whitespace ignored) and `default` when the variable is unset or
#' empty. Mirrors the truthy parsing used by the Python client (pysynthbio).
#' @param name The environment variable name.
#' @param default Logical value to return when the variable is unset/empty.
#' @return A length-one logical.
#' @noRd
env_flag <- function(name, default = FALSE) {
  value <- Sys.getenv(name, unset = NA_character_)
  if (is.na(value) || !nzchar(value)) {
    return(default)
  }
  tolower(trimws(value)) %in% c("1", "true", "yes", "on")
}

# A model's variant slugs (reference-conditioning, predict-metadata) are served
# by the same container as their base model, so they resolve to the same host.
MODEL_ID_SUFFIXES <- c("_reference-conditioning", "_predict-metadata")

#' @title Reduce a Model Slug to its Base Model (Internal)
#' @description Strips the variant suffixes so all slugs backed by one container
#' map to the same per-model environment variable.
#' @param model_id A model slug.
#' @return The base model id.
#' @noRd
base_model_id <- function(model_id) {
  for (suffix in MODEL_ID_SUFFIXES) {
    if (endsWith(model_id, suffix)) {
      return(substr(model_id, 1, nchar(model_id) - nchar(suffix)))
    }
  }
  model_id
}

#' @title Per-Model Base URL Environment Variable Name (Internal)
#' @description Builds the env var that holds the self-hosted base URL for a
#' specific model, e.g. `gem-1-bulk` (and its variants) ->
#' `SYNTHESIZE_API_BASE_URL__GEM_1_BULK`. Mirrors the pysynthbio naming.
#' @param model_id A model slug.
#' @return The environment variable name.
#' @noRd
per_model_env_var <- function(model_id) {
  key <- gsub("[^A-Z0-9]+", "_", toupper(base_model_id(model_id)))
  paste0("SYNTHESIZE_API_BASE_URL__", key)
}

#' @title Resolve the API Base URL (Internal)
#' @description Resolves the base URL in precedence order: an explicit
#' `api_base_url` argument, then the per-model environment variable
#' `SYNTHESIZE_API_BASE_URL__<MODEL>` (when `model_id` is supplied), then the
#' global `SYNTHESIZE_API_BASE_URL`, then the production default. The per-model
#' variable lets a scientist point each model at its own self-hosted container
#' once and never pass a URL on every call.
#' @param api_base_url An explicit base URL, or NULL to resolve from the
#'        environment.
#' @param model_id Optional model slug used to look up a per-model variable.
#' @return A character scalar base URL.
#' @noRd
resolve_api_base_url <- function(api_base_url = NULL, model_id = NULL) {
  if (!is.null(api_base_url)) {
    return(api_base_url)
  }
  if (!is.null(model_id)) {
    per_model <- Sys.getenv(per_model_env_var(model_id), unset = NA_character_)
    if (!is.na(per_model) && nzchar(per_model)) {
      return(per_model)
    }
  }
  Sys.getenv("SYNTHESIZE_API_BASE_URL", unset = API_BASE_URL)
}

#' @title Resolve the Self-Hosted Flag (Internal)
#' @description An explicit `self_hosted` value always wins. Otherwise the flag
#' is read from the `SYNTHESIZE_SELF_HOSTED` environment variable (truthy for
#' 1/true/yes/on), defaulting to FALSE so the production async path is unchanged
#' when neither the argument nor the variable is set.
#' @param self_hosted An explicit logical, or NULL to resolve from the
#'        environment.
#' @return A length-one logical.
#' @noRd
resolve_self_hosted <- function(self_hosted = NULL) {
  if (!is.null(self_hosted)) {
    return(isTRUE(self_hosted))
  }
  env_flag("SYNTHESIZE_SELF_HOSTED")
}

library(testthat)
library(mockery)
library(jsonlite)

# -----------------------------------------------------------------------------
# Helpers: build in-memory Arrow IPC stream fixtures that mimic the self-hosted
# container's synchronous responses (schema-level metadata + per-row columns).
# -----------------------------------------------------------------------------

make_baseline_stream <- function(gene_order,
                                 counts_list,
                                 metadata_df,
                                 request_type = "baseline",
                                 model_version = "gem-1-bulk-v2",
                                 with_latents = TRUE) {
  n <- length(counts_list)
  columns <- list(
    counts = counts_list,
    metadata = metadata_df
  )
  if (with_latents) {
    columns$biological_latent <- replicate(n, c(0.1, 0.2, 0.3), simplify = FALSE)
    columns$technical_latent <- replicate(n, c(1.1, 1.2), simplify = FALSE)
    columns$perturbation_latent <- replicate(n, c(2.1, 2.2, 2.3, 2.4), simplify = FALSE)
  }

  tab <- do.call(arrow::arrow_table, columns)
  tab <- tab$ReplaceSchemaMetadata(list(
    model_version = model_version,
    request_type = request_type,
    gene_order = as.character(jsonlite::toJSON(gene_order))
  ))
  arrow::write_to_raw(tab, format = "stream")
}

make_metadata_stream <- function(counts_list,
                                 metadata_df,
                                 model_version = "gem-1-bulk-predict-metadata-v2") {
  n <- length(counts_list)
  latents_df <- data.frame(
    biological = I(replicate(n, c(0.1, 0.2), simplify = FALSE)),
    technical = I(replicate(n, c(1.1, 1.2), simplify = FALSE)),
    perturbation = I(replicate(n, c(2.1, 2.2), simplify = FALSE))
  )
  classifier_df <- data.frame(
    sex = I(data.frame(
      male = runif(n),
      female = runif(n)
    ))
  )
  decoder_df <- data.frame(counts = I(counts_list))

  tab <- arrow::arrow_table(
    classifier_probs = classifier_df,
    latents = latents_df,
    metadata = metadata_df,
    decoder_sample = decoder_df
  )
  tab <- tab$ReplaceSchemaMetadata(list(
    model_version = model_version,
    request_type = "predict_metadata"
  ))
  arrow::write_to_raw(tab, format = "stream")
}

# -----------------------------------------------------------------------------
# Baseline / reference_conditioned Arrow parsing
# -----------------------------------------------------------------------------

test_that("transform_arrow_table parses a baseline response (mocked Arrow stream)", {
  skip_if_not_installed("arrow")

  gene_order <- c("ENSG00000000003", "ENSG00000000005", "ENSG00000000419")
  counts_list <- list(
    as.integer(c(100, 200, 300)),
    as.integer(c(150, 250, 350))
  )
  metadata_df <- data.frame(
    sample_type = c("cell line", "primary tissue"),
    sex = c("male", "female"),
    stringsAsFactors = FALSE
  )

  raw_bytes <- make_baseline_stream(gene_order, counts_list, metadata_df)
  table <- rsynthbio:::read_arrow_stream(raw_bytes)
  result <- rsynthbio:::transform_arrow_table(table)

  expect_named(result, c("metadata", "expression", "latents"))

  # Expression: gene_order column names and correct ordering of counts
  expect_s3_class(result$expression, "data.frame")
  expect_equal(colnames(result$expression), gene_order)
  expect_equal(nrow(result$expression), 2)
  expect_equal(as.numeric(result$expression[1, ]), c(100, 200, 300))
  expect_equal(as.numeric(result$expression[2, ]), c(150, 250, 350))

  # Metadata: sample_id prepended, original columns preserved
  expect_s3_class(result$metadata, "data.frame")
  expect_equal(nrow(result$metadata), 2)
  expect_true("sample_id" %in% colnames(result$metadata))
  expect_equal(result$metadata$sex, c("male", "female"))

  # Latents reshaped to biological/technical/perturbation list columns
  expect_s3_class(result$latents, "data.frame")
  expect_named(result$latents, c("biological", "technical", "perturbation"))
  expect_equal(length(result$latents$biological[[1]]), 3)

  # Schema metadata exposed as attributes (shape unchanged)
  expect_equal(attr(result, "model_version"), "gem-1-bulk-v2")
  expect_equal(attr(result, "request_type"), "baseline")
})

test_that("transform_arrow_table handles reference_conditioned the same as baseline", {
  skip_if_not_installed("arrow")

  gene_order <- c("gene1", "gene2", "gene3")
  raw_bytes <- make_baseline_stream(
    gene_order,
    counts_list = list(as.integer(c(1, 2, 3))),
    metadata_df = data.frame(sample_type = "cell line", stringsAsFactors = FALSE),
    request_type = "reference_conditioned"
  )
  result <- rsynthbio:::transform_arrow_table(rsynthbio:::read_arrow_stream(raw_bytes))

  expect_equal(colnames(result$expression), gene_order)
  expect_equal(attr(result, "request_type"), "reference_conditioned")
})

test_that("transform_arrow_table omits latents when absent", {
  skip_if_not_installed("arrow")

  raw_bytes <- make_baseline_stream(
    gene_order = c("g1", "g2"),
    counts_list = list(as.integer(c(5, 6))),
    metadata_df = data.frame(sex = "male", stringsAsFactors = FALSE),
    with_latents = FALSE
  )
  result <- rsynthbio:::transform_arrow_table(rsynthbio:::read_arrow_stream(raw_bytes))

  expect_named(result, c("metadata", "expression"))
})

# -----------------------------------------------------------------------------
# predict_metadata Arrow parsing
# -----------------------------------------------------------------------------

test_that("transform_arrow_table parses a predict_metadata response", {
  skip_if_not_installed("arrow")

  counts_list <- list(
    as.integer(c(10, 20, 30)),
    as.integer(c(40, 50, 60))
  )
  metadata_df <- data.frame(
    sex = c("male", "female"),
    sample_type = c("cell line", "primary tissue"),
    stringsAsFactors = FALSE
  )

  raw_bytes <- make_metadata_stream(counts_list, metadata_df)
  result <- rsynthbio:::transform_arrow_table(rsynthbio:::read_arrow_stream(raw_bytes))

  expect_named(result, c("metadata", "latents", "classifier_probs", "expression"))
  expect_s3_class(result$expression, "data.frame")
  expect_equal(nrow(result$expression), 2)
  expect_equal(ncol(result$expression), 3)
  expect_equal(as.numeric(result$expression[1, ]), c(10, 20, 30))
  expect_s3_class(result$metadata, "data.frame")
  expect_equal(nrow(result$metadata), 2)
  expect_equal(attr(result, "request_type"), "predict_metadata")
})

# -----------------------------------------------------------------------------
# Error handling
# -----------------------------------------------------------------------------

test_that("transform_arrow_table errors on missing request_type metadata", {
  skip_if_not_installed("arrow")

  tab <- arrow::arrow_table(
    counts = list(as.integer(c(1, 2))),
    metadata = data.frame(sex = "male", stringsAsFactors = FALSE)
  )
  raw_bytes <- arrow::write_to_raw(tab, format = "stream")
  table <- rsynthbio:::read_arrow_stream(raw_bytes)

  expect_error(
    rsynthbio:::transform_arrow_table(table),
    "missing required 'request_type'"
  )
})

test_that("transform_arrow_table errors on unsupported request_type", {
  skip_if_not_installed("arrow")

  tab <- arrow::arrow_table(
    counts = list(as.integer(c(1, 2))),
    metadata = data.frame(sex = "male", stringsAsFactors = FALSE)
  )
  tab <- tab$ReplaceSchemaMetadata(list(request_type = "something_new"))
  raw_bytes <- arrow::write_to_raw(tab, format = "stream")

  expect_error(
    rsynthbio:::transform_arrow_table(rsynthbio:::read_arrow_stream(raw_bytes)),
    "Unsupported request_type.*something_new"
  )
})

test_that("require_arrow errors with an install hint when arrow is missing", {
  require_arrow_fn <- rsynthbio:::require_arrow
  stub(require_arrow_fn, "requireNamespace", function(...) FALSE)

  expect_error(
    require_arrow_fn(),
    "install.packages\\(\"arrow\"\\)"
  )
})

# -----------------------------------------------------------------------------
# predict_query_self_hosted end-to-end (HTTP mocked, real Arrow parsing)
# -----------------------------------------------------------------------------

test_that("predict_query_self_hosted parses a baseline stream without an API key", {
  skip_if_not_installed("arrow")

  original_api_key <- Sys.getenv("SYNTHESIZE_API_KEY")
  Sys.unsetenv("SYNTHESIZE_API_KEY")
  on.exit({
    if (nzchar(original_api_key)) Sys.setenv(SYNTHESIZE_API_KEY = original_api_key)
  })

  gene_order <- c("g1", "g2", "g3")
  raw_bytes <- make_baseline_stream(
    gene_order,
    counts_list = list(as.integer(c(7, 8, 9))),
    metadata_df = data.frame(sex = "male", stringsAsFactors = FALSE)
  )

  fn <- rsynthbio:::predict_query_self_hosted
  stub(fn, "POST", structure(list(status_code = 200), class = "response"))
  stub(fn, "status_code", function(x) 200)
  stub(fn, "content", function(x, as) raw_bytes)

  result <- fn(
    query = list(inputs = list()),
    model_id = "gem-1-bulk",
    api_base_url = "https://gem-1-bulk.internal.partner.example",
    raw_response = FALSE
  )

  expect_equal(colnames(result$expression), gene_order)
  expect_equal(as.numeric(result$expression[1, ]), c(7, 8, 9))
})

test_that("predict_query_self_hosted returns the raw Arrow table when requested", {
  skip_if_not_installed("arrow")

  gene_order <- c("g1", "g2")
  raw_bytes <- make_baseline_stream(
    gene_order,
    counts_list = list(as.integer(c(3, 4))),
    metadata_df = data.frame(sex = "female", stringsAsFactors = FALSE)
  )

  fn <- rsynthbio:::predict_query_self_hosted
  stub(fn, "POST", structure(list(status_code = 200), class = "response"))
  stub(fn, "status_code", function(x) 200)
  stub(fn, "content", function(x, as) raw_bytes)

  result <- fn(
    query = list(inputs = list()),
    model_id = "gem-1-bulk",
    api_base_url = "https://host",
    raw_response = TRUE
  )

  expect_equal(result$request_type, "baseline")
  expect_equal(result$gene_order, gene_order)
  expect_true(inherits(result$table, "Table"))
})

test_that("predict_query_self_hosted raises on HTTP errors", {
  skip_if_not_installed("arrow")

  fn <- rsynthbio:::predict_query_self_hosted
  stub(fn, "POST", structure(list(status_code = 500), class = "response"))
  stub(fn, "status_code", function(x) 500)
  stub(fn, "content", function(x, ...) "Internal Server Error")

  expect_error(
    fn(
      query = list(inputs = list()),
      model_id = "gem-1-bulk",
      api_base_url = "https://host",
      raw_response = FALSE
    ),
    "Self-hosted predict request failed with status 500"
  )
})

test_that("predict_query dispatches to the self-hosted path when self_hosted = TRUE", {
  dispatch <- mockery::mock(list(dispatched = TRUE))
  stub(predict_query, "predict_query_self_hosted", dispatch)

  result <- predict_query(
    query = list(inputs = list()),
    model_id = "gem-1-bulk",
    api_base_url = "https://host",
    self_hosted = TRUE
  )

  expect_true(result$dispatched)
  expect_called(dispatch, 1)
})

# -----------------------------------------------------------------------------
# make_api_request self-hosted (no token) behaviour
# -----------------------------------------------------------------------------

test_that("make_api_request does not require a token in self_hosted mode", {
  original_api_key <- Sys.getenv("SYNTHESIZE_API_KEY")
  Sys.unsetenv("SYNTHESIZE_API_KEY")
  on.exit({
    if (nzchar(original_api_key)) Sys.setenv(SYNTHESIZE_API_KEY = original_api_key)
  })

  fn <- rsynthbio:::make_api_request
  stub(fn, "GET", structure(list(status_code = 200), class = "response"))
  stub(fn, "status_code", function(x) 200)
  stub(fn, "content", function(x, type) '{"ok":true}')

  result <- fn("https://host/api/models", "List models", self_hosted = TRUE)
  expect_true(result$ok)
})

test_that("make_api_request still requires a token in the production path", {
  original_api_key <- Sys.getenv("SYNTHESIZE_API_KEY")
  Sys.unsetenv("SYNTHESIZE_API_KEY")
  on.exit({
    if (nzchar(original_api_key)) Sys.setenv(SYNTHESIZE_API_KEY = original_api_key)
  })

  expect_error(
    rsynthbio:::make_api_request("https://host/api/models", "List models"),
    "set your API key"
  )
})

# -----------------------------------------------------------------------------
# Environment-variable defaults (parity with the Python client)
# -----------------------------------------------------------------------------

test_that("env_flag parses truthy values case-insensitively", {
  for (truthy in c("1", "true", "TRUE", "Yes", " on ", "ON")) {
    Sys.setenv(RSYNTHBIO_TEST_FLAG = truthy)
    expect_true(rsynthbio:::env_flag("RSYNTHBIO_TEST_FLAG"), info = truthy)
  }
  for (falsy in c("0", "false", "no", "off", "", "nope")) {
    Sys.setenv(RSYNTHBIO_TEST_FLAG = falsy)
    expect_false(rsynthbio:::env_flag("RSYNTHBIO_TEST_FLAG"), info = falsy)
  }
  Sys.unsetenv("RSYNTHBIO_TEST_FLAG")
  expect_false(rsynthbio:::env_flag("RSYNTHBIO_TEST_FLAG"))
  expect_true(rsynthbio:::env_flag("RSYNTHBIO_TEST_FLAG", default = TRUE))
})

test_that("resolve_api_base_url prefers explicit arg, then env var, then default", {
  original <- Sys.getenv("SYNTHESIZE_API_BASE_URL", unset = NA_character_)
  on.exit({
    if (is.na(original)) Sys.unsetenv("SYNTHESIZE_API_BASE_URL") else Sys.setenv(SYNTHESIZE_API_BASE_URL = original)
  })

  Sys.setenv(SYNTHESIZE_API_BASE_URL = "https://env.example")
  expect_equal(rsynthbio:::resolve_api_base_url("https://explicit.example"), "https://explicit.example")
  expect_equal(rsynthbio:::resolve_api_base_url(NULL), "https://env.example")

  Sys.unsetenv("SYNTHESIZE_API_BASE_URL")
  expect_equal(rsynthbio:::resolve_api_base_url(NULL), API_BASE_URL)
})

test_that("per_model_env_var builds the model-specific variable name", {
  expect_equal(rsynthbio:::per_model_env_var("gem-1-bulk"), "SYNTHESIZE_API_BASE_URL__GEM_1_BULK")
  expect_equal(rsynthbio:::per_model_env_var("gem-1-sc"), "SYNTHESIZE_API_BASE_URL__GEM_1_SC")
  # Variant slugs normalize to their base model's variable.
  expect_equal(
    rsynthbio:::per_model_env_var("gem-1-bulk_predict-metadata"),
    "SYNTHESIZE_API_BASE_URL__GEM_1_BULK"
  )
  expect_equal(
    rsynthbio:::per_model_env_var("gem-1-sc_reference-conditioning"),
    "SYNTHESIZE_API_BASE_URL__GEM_1_SC"
  )
})

test_that("resolve_api_base_url routes per-model without per-call URLs", {
  saved <- c("SYNTHESIZE_API_BASE_URL", "SYNTHESIZE_API_BASE_URL__GEM_1_BULK", "SYNTHESIZE_API_BASE_URL__GEM_1_SC")
  originals <- vapply(saved, function(v) Sys.getenv(v, unset = NA_character_), character(1))
  on.exit({
    for (v in saved) {
      if (is.na(originals[[v]])) Sys.unsetenv(v) else do.call(Sys.setenv, setNames(list(originals[[v]]), v))
    }
  })

  Sys.unsetenv("SYNTHESIZE_API_BASE_URL")
  Sys.setenv(SYNTHESIZE_API_BASE_URL__GEM_1_BULK = "http://bulk:8080")
  Sys.setenv(SYNTHESIZE_API_BASE_URL__GEM_1_SC = "http://sc:8080")

  # Each model resolves to its own host.
  expect_equal(rsynthbio:::resolve_api_base_url(model_id = "gem-1-bulk"), "http://bulk:8080")
  expect_equal(rsynthbio:::resolve_api_base_url(model_id = "gem-1-sc"), "http://sc:8080")
  # Variant slugs share the base model's host.
  expect_equal(rsynthbio:::resolve_api_base_url(model_id = "gem-1-bulk_predict-metadata"), "http://bulk:8080")
  expect_equal(rsynthbio:::resolve_api_base_url(model_id = "gem-1-sc_reference-conditioning"), "http://sc:8080")
  # Explicit arg still wins over the per-model env var.
  expect_equal(
    rsynthbio:::resolve_api_base_url("http://explicit:9000", model_id = "gem-1-bulk"),
    "http://explicit:9000"
  )

  # Per-model wins over the global, but a model without its own var uses the global.
  Sys.setenv(SYNTHESIZE_API_BASE_URL = "http://global:8080")
  expect_equal(rsynthbio:::resolve_api_base_url(model_id = "gem-1-sc"), "http://sc:8080")
  Sys.unsetenv("SYNTHESIZE_API_BASE_URL__GEM_1_BULK")
  expect_equal(rsynthbio:::resolve_api_base_url(model_id = "gem-1-bulk"), "http://global:8080")
})

test_that("resolve_self_hosted prefers explicit arg over env var", {
  original <- Sys.getenv("SYNTHESIZE_SELF_HOSTED", unset = NA_character_)
  on.exit({
    if (is.na(original)) Sys.unsetenv("SYNTHESIZE_SELF_HOSTED") else Sys.setenv(SYNTHESIZE_SELF_HOSTED = original)
  })

  Sys.setenv(SYNTHESIZE_SELF_HOSTED = "true")
  expect_false(rsynthbio:::resolve_self_hosted(FALSE))
  expect_true(rsynthbio:::resolve_self_hosted(NULL))

  Sys.setenv(SYNTHESIZE_SELF_HOSTED = "off")
  expect_false(rsynthbio:::resolve_self_hosted(NULL))
  expect_true(rsynthbio:::resolve_self_hosted(TRUE))

  Sys.unsetenv("SYNTHESIZE_SELF_HOSTED")
  expect_false(rsynthbio:::resolve_self_hosted(NULL))
})

test_that("predict_query enables self-hosted from SYNTHESIZE_SELF_HOSTED env var", {
  original_flag <- Sys.getenv("SYNTHESIZE_SELF_HOSTED", unset = NA_character_)
  original_url <- Sys.getenv("SYNTHESIZE_API_BASE_URL", unset = NA_character_)
  on.exit({
    if (is.na(original_flag)) Sys.unsetenv("SYNTHESIZE_SELF_HOSTED") else Sys.setenv(SYNTHESIZE_SELF_HOSTED = original_flag)
    if (is.na(original_url)) Sys.unsetenv("SYNTHESIZE_API_BASE_URL") else Sys.setenv(SYNTHESIZE_API_BASE_URL = original_url)
  })

  Sys.setenv(SYNTHESIZE_SELF_HOSTED = "1")
  Sys.setenv(SYNTHESIZE_API_BASE_URL = "https://gem-1-bulk.internal.partner.example")

  dispatch <- mockery::mock(list(dispatched = TRUE))
  stub(predict_query, "predict_query_self_hosted", dispatch)

  result <- predict_query(query = list(inputs = list()), model_id = "gem-1-bulk")

  expect_true(result$dispatched)
  expect_called(dispatch, 1)
  # The resolved base URL from the env var is forwarded to the self-hosted path.
  call_args <- mockery::mock_args(dispatch)[[1]]
  expect_equal(call_args$api_base_url, "https://gem-1-bulk.internal.partner.example")
})

test_that("list_models resolves base URL from the env var when not supplied", {
  original_url <- Sys.getenv("SYNTHESIZE_API_BASE_URL", unset = NA_character_)
  on.exit({
    if (is.na(original_url)) Sys.unsetenv("SYNTHESIZE_API_BASE_URL") else Sys.setenv(SYNTHESIZE_API_BASE_URL = original_url)
  })
  Sys.setenv(SYNTHESIZE_API_BASE_URL = "https://gem-1-bulk.internal.partner.example")

  seen_url <- NULL
  stub(list_models, "make_api_request", function(url, context_msg, ...) {
    seen_url <<- url
    list(ok = TRUE)
  })

  list_models()
  expect_equal(seen_url, "https://gem-1-bulk.internal.partner.example/api/models")
})

# Helper functions for mocking API calls in tests

#' Setup temporary API key for testing
#'
#' @param api_key The API key to set (default: "mock-api-key")
#' @return The original API key value (to be used in cleanup)
setup_mock_api_key <- function(api_key = "mock-api-key") {
    original_api_key <- Sys.getenv("SYNTHESIZE_API_KEY")
    Sys.setenv(SYNTHESIZE_API_KEY = api_key)
    original_api_key
}

#' Restore original API key
#'
#' @param original_api_key The original API key to restore
restore_api_key <- function(original_api_key) {
    if (original_api_key != "") {
        Sys.setenv(SYNTHESIZE_API_KEY = original_api_key)
    } else {
        Sys.unsetenv("SYNTHESIZE_API_KEY")
    }
}

#' Create standard mocks for a successful query
#'
#' @param query_id The query ID to return from start_model_query
#' @param download_url The download URL to return from poll
#' @param counts_list A list of count vectors, one per sample
#' @param metadata_df A data frame of metadata, one row per sample
#' @param gene_order The gene order to return from get_json
#' @return A list of mocks: has_token, start_query, poll, get_json
create_success_mocks <- function(query_id = "test-id",
                                 download_url = "https://example.com/data.json",
                                 counts_list = list(c(100, 200, 300)),
                                 metadata_df = data.frame(sample_id = "test1"),
                                 gene_order = c("gene1", "gene2", "gene3")) {
    list(
        has_token = mockery::mock(TRUE, cycle = TRUE),
        start_query = mockery::mock(query_id, cycle = TRUE),
        poll = mockery::mock(
            list(
                status = "ready",
                payload = list(
                    status = "ready",
                    downloadUrl = download_url
                )
            ),
            cycle = TRUE
        ),
        get_json = mockery::mock(
            list(
                outputs = list(
                    counts = counts_list,
                    metadata = metadata_df
                ),
                gene_order = gene_order,
                model_version = 2
            ),
            cycle = TRUE
        )
    )
}

#' Create mocks for a failed query
#'
#' @param query_id The query ID to return from start_model_query
#' @param error_message The error message to return
#' @return A list of mocks: has_token, start_query, poll
create_failure_mocks <- function(query_id = "test-id",
                                 error_message = "Query failed") {
    list(
        has_token = mockery::mock(TRUE, cycle = TRUE),
        start_query = mockery::mock(query_id, cycle = TRUE),
        poll = mockery::mock(
            list(
                status = "failed",
                payload = list(
                    status = "failed",
                    message = error_message
                )
            ),
            cycle = TRUE
        )
    )
}

#' Create mocks for a timeout scenario
#'
#' @param query_id The query ID to return from start_model_query
#' @return A list of mocks: has_token, start_query, poll
create_timeout_mocks <- function(query_id = "test-id") {
    list(
        has_token = mockery::mock(TRUE, cycle = TRUE),
        start_query = mockery::mock(query_id, cycle = TRUE),
        poll = mockery::mock(
            list(
                status = "running",
                payload = list(status = "running")
            ),
            cycle = TRUE
        )
    )
}

#' Helper to setup environment for a test
#'
#' @param api_key The API key to set
#' @return The original API key (for cleanup)
setup_test_environment <- function(api_key = "mock-api-key") {
    setup_mock_api_key(api_key)
}

library(testthat)
library(mockery)
library(jsonlite)
library(httr)

# -----------------------------
# Mocked Async Tests (Bulk)
# -----------------------------

test_that("predict_query mocked call success (bulk async)", {
  # Setup environment
  original_api_key <- setup_test_environment("mock-api-key-for-test")
  on.exit(restore_api_key(original_api_key))

  # Create mocks for successful query
  mocks <- create_success_mocks(
    query_id = "bulk-xyz",
    download_url = "https://example.com/bulk.json",
    counts_list = list(
      c(100, 200, 300),
      c(150, 250, 350)
    ),
    metadata_df = data.frame(
      sample_id = c("test1", "test2"),
      cell_line_ontology_id = c("CVCL_0023", "CVCL_0023"),
      age_years = c("25", "30"),
      sex = c("female", "male")
    ),
    gene_order = c("gene1", "gene2", "gene3")
  )

  # Apply stubs directly
  stub(predict_query, "has_synthesize_token", mocks$has_token)
  stub(predict_query, "start_model_query", mocks$start_query)
  stub(predict_query, "poll_model_query", mocks$poll)
  stub(predict_query, "get_json", mocks$get_json)

  test_query <- list(
    inputs = list(
      list(
        metadata = list(
          cell_type_ontology_id = "CL:0000786",
          tissue_ontology_id = "UBERON:0001155",
          sex = "male",
          sample_type = "primary tissue"
        ),
        num_samples = 5
      )
    ),
    mode = "mean estimation",
    seed = 11
  )
  results <- predict_query(query = test_query, model_id = "gem-1-bulk")

  expect_type(results, "list")
  expect_true("metadata" %in% names(results))
  expect_true("expression" %in% names(results))

  expect_s3_class(results$metadata, "data.frame")
  expect_s3_class(results$expression, "data.frame")

  expect_equal(nrow(results$metadata), 2)
  expect_equal(nrow(results$expression), 2)
  expect_equal(ncol(results$expression), 3)

  # Check data values
  expect_equal(as.numeric(results$expression[1, ]), c(100, 200, 300))
  expect_equal(as.numeric(results$expression[2, ]), c(150, 250, 350))

  expect_called(mocks$start_query, 1)
  expect_called(mocks$poll, 1)
  expect_called(mocks$get_json, 1)
})

test_that("predict_query new API structure handling (bulk)", {
  original_api_key <- setup_test_environment()
  on.exit(restore_api_key(original_api_key))

  # Create large gene order
  gene_order <- paste0("ENSG", sprintf("%011d", 1:44590))

  mocks <- create_success_mocks(
    query_id = "bulk-big",
    download_url = "https://example.com/big.json",
    counts_list = list(
      rep(c(0.1, 0.2, 0.3, 0.4, 0.5), 8918)
    ),
    metadata_df = data.frame(
      age_years = "65",
      disease_ontology_id = "MONDO:0011719",
      sex = "female",
      sample_type = "primary tissue",
      tissue_ontology_id = "UBERON:0000945"
    ),
    gene_order = gene_order
  )

  stub(predict_query, "has_synthesize_token", mocks$has_token)
  stub(predict_query, "start_model_query", mocks$start_query)
  stub(predict_query, "poll_model_query", mocks$poll)
  stub(predict_query, "get_json", mocks$get_json)

  query <- list(
    inputs = list(
      list(
        metadata = list(
          age_years = "65",
          disease_ontology_id = "MONDO:0011719",
          sex = "female",
          sample_type = "primary tissue",
          tissue_ontology_id = "UBERON:0000945"
        ),
        num_samples = 1
      )
    ),
    mode = "mean estimation",
    seed = 11
  )
  results <- predict_query(query, model_id = "gem-1-bulk")

  expect_true("metadata" %in% names(results))
  expect_true("expression" %in% names(results))
  expect_equal(nrow(results$metadata), 1)
  expect_equal(nrow(results$expression), 1)
  expect_equal(ncol(results$expression), 44590)
})

# -----------------------------
# Mocked Single-Cell Async Tests
# -----------------------------

test_that("predict_query single-cell success (mocked)", {
  original_api_key <- setup_test_environment("test-api-token")
  on.exit(restore_api_key(original_api_key))

  mocks <- create_success_mocks(
    query_id = "abc123",
    download_url = "https://example.com/final.json",
    counts_list = list(
      c(1, 2, 3),
      c(4, 5, 6)
    ),
    metadata_df = data.frame(
      sample_id = c("s1", "s2")
    ),
    gene_order = c("gene1", "gene2", "gene3")
  )

  stub(predict_query, "has_synthesize_token", mocks$has_token)
  stub(predict_query, "start_model_query", mocks$start_query)
  stub(predict_query, "poll_model_query", mocks$poll)
  stub(predict_query, "get_json", mocks$get_json)

  query <- list(
    inputs = list(
      list(
        metadata = list(
          cell_type_ontology_id = "CL:0000786",
          tissue_ontology_id = "UBERON:0001155",
          sex = "male"
        ),
        num_samples = 5
      )
    ),
    mode = "mean estimation",
    seed = 11
  )
  result <- predict_query(query, model_id = "gem-1-sc")

  expect_true("metadata" %in% names(result))
  expect_true("expression" %in% names(result))
  expect_equal(nrow(result$metadata), 2)
  expect_equal(colnames(result$expression), c("gene1", "gene2", "gene3"))
  expect_equal(as.numeric(result$expression[1, ]), c(1, 2, 3))
  expect_equal(as.numeric(result$expression[2, ]), c(4, 5, 6))
})

test_that("predict_query single-cell failure (mocked)", {
  original_api_key <- setup_test_environment("test-api-token")
  on.exit(restore_api_key(original_api_key))

  mocks <- create_failure_mocks(
    query_id = "abc123",
    error_message = "Invalid metadata: missing required field 'cell_type_ontology_id'"
  )

  stub(predict_query, "has_synthesize_token", mocks$has_token)
  stub(predict_query, "start_model_query", mocks$start_query)
  stub(predict_query, "poll_model_query", mocks$poll)

  query <- list(
    inputs = list(
      list(
        metadata = list(
          cell_type_ontology_id = "CL:0000786",
          tissue_ontology_id = "UBERON:0001155",
          sex = "male"
        ),
        num_samples = 5
      )
    ),
    mode = "mean estimation",
    seed = 11
  )

  expect_error(
    predict_query(query, model_id = "gem-1-sc"),
    "Model query failed.*missing required field"
  )
})

test_that("predict_query single-cell timeout (mocked)", {
  original_api_key <- setup_test_environment("test-api-token")
  on.exit(restore_api_key(original_api_key))

  mocks <- create_timeout_mocks(query_id = "abc123")

  stub(predict_query, "has_synthesize_token", mocks$has_token)
  stub(predict_query, "start_model_query", mocks$start_query)
  stub(predict_query, "poll_model_query", mocks$poll)

  query <- list(
    inputs = list(
      list(
        metadata = list(
          cell_type_ontology_id = "CL:0000786",
          tissue_ontology_id = "UBERON:0001155",
          sex = "male"
        ),
        num_samples = 5
      )
    ),
    mode = "mean estimation",
    seed = 11
  )

  expect_error(
    predict_query(
      query,
      model_id = "gem-1-sc",
      poll_interval_seconds = 0,
      poll_timeout_seconds = 0
    ),
    "did not complete in time"
  )
})

# -----------------------------
# New API Function Tests
# -----------------------------

test_that("list_models mock test", {
  original_api_key <- setup_test_environment()
  on.exit(restore_api_key(original_api_key))

  # Stub make_api_request to return mock data
  stub(list_models, "make_api_request", function(url, context_msg, ...) {
    list(
      list(id = "gem-1-bulk", name = "GEM-1 Bulk"),
      list(id = "gem-1-sc", name = "GEM-1 Single Cell")
    )
  })

  result <- list_models()
  expect_type(result, "list")
  expect_true(length(result) > 0)
})

test_that("get_example_query mock test", {
  original_api_key <- setup_test_environment()
  on.exit(restore_api_key(original_api_key))

  # Stub make_api_request to return mock data
  stub(get_example_query, "make_api_request", function(url, context_msg, ...) {
    list(
      inputs = list(
        list(
          metadata = list(cell_type_ontology_id = "CL:0000786"),
          num_samples = 5
        )
      ),
      mode = "mean estimation"
    )
  })

  result <- get_example_query(model_id = "gem-1-bulk")
  expect_type(result, "list")
  expect_true("inputs" %in% names(result))
  expect_true("mode" %in% names(result))
})

# -----------------------------
# get_json Tests
# -----------------------------

test_that("get_json successfully fetches and parses JSON", {
  test_data <- list(
    outputs = list(
      list(counts = c(1, 2, 3), metadata = list(sample_id = "test"))
    ),
    gene_order = c("gene1", "gene2", "gene3")
  )

  # Mock successful HTTP response
  mock_response <- mock(
    structure(
      list(
        url = "https://example.com/data.json",
        status_code = 200,
        content = charToRaw(toJSON(test_data, auto_unbox = TRUE))
      ),
      class = "response"
    ),
    cycle = TRUE
  )

  get_json_fn <- rsynthbio:::get_json
  stub(get_json_fn, "GET", mock_response)
  stub(get_json_fn, "status_code", function(x) 200)
  stub(get_json_fn, "content", function(x, type, encoding) {
    toJSON(test_data, auto_unbox = TRUE)
  })

  result <- get_json_fn("https://example.com/data.json")

  expect_type(result, "list")
  expect_true("outputs" %in% names(result))
  expect_true("gene_order" %in% names(result))
  expect_equal(result$gene_order, c("gene1", "gene2", "gene3"))
})

test_that("get_json handles network errors", {
  # Mock network error
  mock_get_error <- mock(stop("Connection timeout"), cycle = TRUE)

  get_json_fn <- rsynthbio:::get_json
  stub(get_json_fn, "GET", mock_get_error)

  expect_error(
    get_json_fn("https://example.com/data.json"),
    "Failed to fetch download URL due to a network issue.*Connection timeout"
  )
})

test_that("get_json handles HTTP 404 errors", {
  mock_response <- mock(
    structure(
      list(
        url = "https://example.com/data.json",
        status_code = 404,
        content = charToRaw("Not Found")
      ),
      class = "response"
    ),
    cycle = TRUE
  )

  get_json_fn <- rsynthbio:::get_json
  stub(get_json_fn, "GET", mock_response)
  stub(get_json_fn, "status_code", function(x) 404)
  stub(get_json_fn, "content", function(x, type) "Not Found")

  expect_error(
    get_json_fn("https://example.com/data.json"),
    "Download URL fetch failed with status 404.*Not Found"
  )
})

test_that("get_json handles HTTP 500 errors", {
  mock_response <- mock(
    structure(
      list(
        url = "https://example.com/data.json",
        status_code = 500,
        content = charToRaw("Internal Server Error")
      ),
      class = "response"
    ),
    cycle = TRUE
  )

  get_json_fn <- rsynthbio:::get_json
  stub(get_json_fn, "GET", mock_response)
  stub(get_json_fn, "status_code", function(x) 500)
  stub(get_json_fn, "content", function(x, type) "Internal Server Error")

  expect_error(
    get_json_fn("https://example.com/data.json"),
    "Download URL fetch failed with status 500.*Internal Server Error"
  )
})

test_that("get_json handles JSON parsing errors", {
  mock_response <- mock(
    structure(
      list(
        url = "https://example.com/data.json",
        status_code = 200,
        content = charToRaw("invalid json content {{{")
      ),
      class = "response"
    ),
    cycle = TRUE
  )

  get_json_fn <- rsynthbio:::get_json
  stub(get_json_fn, "GET", mock_response)
  stub(get_json_fn, "status_code", function(x) 200)
  stub(get_json_fn, "content", function(x, type, encoding) "invalid json content {{{")
  stub(get_json_fn, "fromJSON", function(x, ...) stop("JSON parsing error"))

  expect_error(
    get_json_fn("https://example.com/data.json"),
    "Failed to decode JSON from download URL response.*JSON parsing error"
  )
})

test_that("get_json preserves UTF-8 encoding", {
  # Test data with special characters
  test_data <- list(
    outputs = list(
      list(
        counts = c(1, 2, 3),
        metadata = list(
          sample_id = "test-utf8-\u00e9\u00f1\u00fc",
          description = "Test with \u2764\ufe0f"
        )
      )
    ),
    gene_order = c("gene1", "gene2", "gene3")
  )

  # Create JSON string
  json_string <- toJSON(test_data, auto_unbox = TRUE)

  mock_response <- mock(
    structure(
      list(
        url = "https://example.com/data.json",
        status_code = 200,
        content = charToRaw(json_string)
      ),
      class = "response"
    ),
    cycle = TRUE
  )

  get_json_fn <- rsynthbio:::get_json
  stub(get_json_fn, "GET", mock_response)
  stub(get_json_fn, "status_code", function(x) 200)
  stub(get_json_fn, "content", function(x, type, encoding) json_string)

  result <- get_json_fn("https://example.com/data.json")

  expect_true("outputs" %in% names(result))
  expect_true("gene_order" %in% names(result))
  # Verify the data structure was parsed
  expect_true(length(result$outputs) >= 1)
  expect_equal(result$gene_order, c("gene1", "gene2", "gene3"))
})

#' Transform Standard Model Output (Internal)
#'
#' @description Extracts and combines gene expression data from a standard API response
#' (e.g., gem-1-bulk, gem-1-sc models).
#'
#' @param final_json The parsed API response list
#' @return A list with:
#'         - metadata: data.frame containing sample metadata
#'         - expression: data.frame containing combined gene expression data
#'         - latents: data.frame containing embeddings (if requested)
#' @noRd
transform_baseline_output <- function(final_json) {
  # Extract the expression matrices and combine them
  counts_list <- final_json$outputs$counts

  # Handle case where jsonlite converts to data.frame with one column of lists
  # This happens when the server returns an array of arrays
  if (is.data.frame(counts_list) && ncol(counts_list) == 1) {
    counts_list <- counts_list[[1]]
  }

  # Convert each sample's expression vector into a row of the expression matrix
  expression <- do.call(rbind, lapply(counts_list, function(x) as.numeric(x)))
  expression <- as.data.frame(expression)

  # Set gene names as column names
  colnames(expression) <- final_json$gene_order

  # Get metadata dataframe
  metadata <- final_json$outputs$metadata

  # Add sample identifiers to match the expression data
  metadata <- data.frame(
    sample_id = paste0("sample_", seq_len(nrow(metadata))),
    metadata
  )

  # if latents - add those too
  if ("latents" %in% names(final_json$outputs)) {
    final_list <- list(
      metadata = metadata,
      expression = expression,
      latents = final_json$outputs$latents
    )
  } else {
    final_list <- list(
      metadata = metadata,
      expression = expression
    )
  }
  # Return both components
  return(final_list)
}

#' Transform Metadata Model Output (Internal)
#'
#' @description Extracts metadata prediction outputs and converts them to
#' data.frames for metadata, latents, classifier_probs, and expression
#' (from decoder_sample counts).
#'
#' Uses the same direct column-access pattern as transform_baseline_output,
#' since jsonlite with simplifyDataFrame=TRUE converts the outputs array
#' into a columnar structure.
#'
#' @param final_json The parsed API response list
#' @return A list with:
#'         - metadata: data.frame of predicted metadata
#'         - latents: data.frame with biological/technical/perturbation columns
#'         - classifier_probs: data.frame of per-category probability dicts
#'         - expression: data.frame of decoder sample counts
#' @noRd
transform_metadata_output <- function(final_json) {
  outputs <- final_json$outputs

  metadata <- outputs$metadata
  latents <- outputs$latents
  classifier_probs <- outputs$classifier_probs

  # Extract expression counts from decoder_sample.
  # Each output's decoder_sample is {"counts": [...]}.
  # jsonlite may simplify this to a data.frame with a "counts" column,
  # or keep it as a nested list.
  decoder_sample <- outputs$decoder_sample
  if (is.data.frame(decoder_sample)) {
    counts_list <- decoder_sample$counts
  } else if (is.list(decoder_sample)) {
    counts_list <- lapply(decoder_sample, function(x) x$counts)
  } else {
    counts_list <- list()
  }

  # Handle jsonlite wrapping counts in a single-column data.frame
  if (is.data.frame(counts_list) && ncol(counts_list) == 1) {
    counts_list <- counts_list[[1]]
  }

  expression <- do.call(rbind, lapply(counts_list, function(x) as.numeric(x)))
  expression <- as.data.frame(expression)

  gene_order <- final_json$gene_order
  if (!is.null(gene_order)) {
    colnames(expression) <- gene_order
  }

  return(list(
    metadata = metadata,
    latents = latents,
    classifier_probs = classifier_probs,
    expression = expression
  ))
}

#' Output Transformer Registry
#'
#' @description A registry mapping model IDs to their corresponding output transformer functions.
#' @noRd
OUTPUT_TRANSFORMERS <- list(
  "gem-1-bulk" = transform_baseline_output,
  "gem-1-sc" = transform_baseline_output,
  "gem-1-bulk_reference-conditioning" = transform_baseline_output,
  "gem-1-sc_reference-conditioning" = transform_baseline_output,
  "gem-1-bulk_condition-on-sample-ids" = transform_baseline_output,
  "gem-1-bulk_predict-metadata" = transform_metadata_output,
  "gem-1-sc_predict-metadata" = transform_metadata_output
)

#' Get Output Transformer for Model (Internal)
#'
#' @description Looks up the appropriate output transformer for a given model ID.
#' Returns NULL if no transformer is registered.
#'
#' @param model_id Character string specifying the model ID
#' @return A transformer function, or NULL if not registered
#' @noRd
get_output_transformer <- function(model_id) {
  return(OUTPUT_TRANSFORMERS[[model_id]])
}

#' @title Arrow IPC Stream Content Type
#' @description MIME type used by the self-hosted container for synchronous
#' Apache Arrow IPC stream responses.
#' @noRd
ARROW_STREAM_CONTENT_TYPE <- "application/vnd.apache.arrow.stream"

#' @title Ensure the arrow Package is Available (Internal)
#' @description Stops with an actionable message when `self_hosted = TRUE` is
#' requested but the optional `arrow` package is not installed. The production
#' (async JSON) path never calls this, so `arrow` remains a soft dependency.
#' @return Invisibly TRUE when `arrow` is available; otherwise throws an error.
#' @noRd
require_arrow <- function() {
  if (!requireNamespace("arrow", quietly = TRUE)) {
    stop(
      paste0(
        "The 'arrow' package is required when self_hosted = TRUE. ",
        "Install it with install.packages(\"arrow\")."
      ),
      call. = FALSE
    )
  }
  invisible(TRUE)
}

#' @title Build Headers for a Self-Hosted Arrow Request (Internal)
#' @description Requests an Apache Arrow IPC stream and only attaches an
#' `Authorization` header when an API key is configured, so the client works
#' against containers that run with authentication disabled.
#' @return An httr `request` object produced by [httr::add_headers()].
#' @importFrom httr add_headers
#' @noRd
build_arrow_post_headers <- function() {
  api_key <- Sys.getenv("SYNTHESIZE_API_KEY")
  if (nzchar(api_key)) {
    return(add_headers(
      Accept = ARROW_STREAM_CONTENT_TYPE,
      `Content-Type` = "application/json",
      Authorization = paste("Bearer", api_key)
    ))
  }
  add_headers(
    Accept = ARROW_STREAM_CONTENT_TYPE,
    `Content-Type` = "application/json"
  )
}

#' @title Read an Arrow IPC Stream into a Table (Internal)
#' @description Parses raw response bytes as an Apache Arrow IPC *stream* (not a
#' file) and returns an Arrow `Table` so schema-level metadata can be inspected.
#' @param raw_bytes A raw vector containing the Arrow IPC stream.
#' @return An Arrow `Table`.
#' @noRd
read_arrow_stream <- function(raw_bytes) {
  tryCatch(
    arrow::read_ipc_stream(raw_bytes, as_data_frame = FALSE),
    error = function(e) {
      stop(
        paste0("Failed to parse Arrow IPC stream from self-hosted response: ", e$message),
        call. = FALSE
      )
    }
  )
}

#' @title Extract Schema-Level Metadata from an Arrow Table (Internal)
#' @param table An Arrow `Table`.
#' @return A named list of metadata strings (possibly empty).
#' @noRd
arrow_schema_metadata <- function(table) {
  metadata <- table$schema$metadata
  if (is.null(metadata)) list() else metadata
}

#' @title Parse the gene_order Schema Metadata Value (Internal)
#' @description The container encodes `gene_order` as a JSON array of gene-id
#' strings. Returns NULL when absent (e.g. predict_metadata responses).
#' @param raw_value The raw `gene_order` metadata string, or NULL.
#' @return A character vector of gene ids, or NULL.
#' @importFrom jsonlite fromJSON
#' @noRd
parse_gene_order <- function(raw_value) {
  if (is.null(raw_value) || !nzchar(raw_value)) {
    return(NULL)
  }
  as.character(fromJSON(raw_value))
}

#' @title Extract a List Column from an Arrow Table as a Base R List (Internal)
#' @description Columns are pulled individually (rather than converting the whole
#' table) to avoid arrow's R-attribute metadata roundtrip, and coerced to a plain
#' list so downstream transformers see the same shape as the JSON path.
#' @param table An Arrow `Table`.
#' @param name The column name to extract.
#' @return A base R list, or NULL when the column is absent.
#' @noRd
arrow_list_column <- function(table, name) {
  if (!(name %in% names(table))) {
    return(NULL)
  }
  as.list(as.vector(table[[name]]))
}

#' @title Extract a Struct Column from an Arrow Table as a data.frame (Internal)
#' @param table An Arrow `Table`.
#' @param name The column name to extract.
#' @return A base data.frame, or NULL when the column is absent.
#' @noRd
arrow_struct_column <- function(table, name) {
  if (!(name %in% names(table))) {
    return(NULL)
  }
  as.data.frame(as.vector(table[[name]]))
}

#' @title Build a Baseline JSON-Shaped List from an Arrow Table (Internal)
#' @description Reshapes the per-row Arrow columns (`counts`, `metadata`, and the
#' three latent list columns) into the same nested list structure that
#' [transform_baseline_output()] consumes, so the JSON and Arrow paths share one
#' transformer and return an identical shape.
#' @param table An Arrow `Table`.
#' @param gene_order Character vector of gene ids (column order for `counts`).
#' @return A list with `gene_order` and `outputs` (counts, metadata, latents).
#' @noRd
build_baseline_json <- function(table, gene_order) {
  outputs <- list(
    counts = arrow_list_column(table, "counts"),
    metadata = arrow_struct_column(table, "metadata")
  )

  biological <- arrow_list_column(table, "biological_latent")
  technical <- arrow_list_column(table, "technical_latent")
  perturbation <- arrow_list_column(table, "perturbation_latent")

  if (!is.null(biological) && !is.null(technical) && !is.null(perturbation)) {
    outputs$latents <- data.frame(
      biological = I(biological),
      technical = I(technical),
      perturbation = I(perturbation)
    )
  }

  list(gene_order = gene_order, outputs = outputs)
}

#' @title Build a Metadata-Prediction JSON-Shaped List (Internal)
#' @description Reshapes predict_metadata Arrow columns into the structure that
#' [transform_metadata_output()] consumes.
#' @param table An Arrow `Table`.
#' @param gene_order Character vector of gene ids, or NULL.
#' @return A list with `gene_order` and `outputs`.
#' @noRd
build_metadata_json <- function(table, gene_order) {
  list(
    gene_order = gene_order,
    outputs = list(
      metadata = arrow_struct_column(table, "metadata"),
      latents = arrow_struct_column(table, "latents"),
      classifier_probs = arrow_struct_column(table, "classifier_probs"),
      decoder_sample = arrow_struct_column(table, "decoder_sample")
    )
  )
}

#' @title Transform a Parsed Arrow Table into Output Data Frames (Internal)
#' @description Reads `request_type`, `model_version`, and `gene_order` from the
#' Arrow schema metadata and dispatches to the matching existing transformer so
#' the return value matches the production JSON path exactly. `model_version` and
#' `request_type` are attached as attributes on the result.
#' @param table An Arrow `Table`.
#' @return A named list of data.frames (see [transform_baseline_output()] /
#'         [transform_metadata_output()]).
#' @noRd
transform_arrow_table <- function(table) {
  metadata <- arrow_schema_metadata(table)
  request_type <- metadata$request_type
  if (is.null(request_type) || !nzchar(request_type)) {
    stop("Arrow response schema is missing required 'request_type' metadata.", call. = FALSE)
  }

  gene_order <- parse_gene_order(metadata$gene_order)

  result <- switch(request_type,
    baseline = transform_baseline_output(build_baseline_json(table, gene_order)),
    reference_conditioned = transform_baseline_output(build_baseline_json(table, gene_order)),
    predict_metadata = transform_metadata_output(build_metadata_json(table, gene_order)),
    stop(paste0("Unsupported request_type in Arrow response: ", request_type), call. = FALSE)
  )

  attr(result, "model_version") <- metadata$model_version
  attr(result, "request_type") <- request_type
  result
}

#' @title Build the Raw (Unformatted) Self-Hosted Result (Internal)
#' @description Returns the parsed Arrow `Table` alongside the schema metadata
#' for callers that pass `raw_response = TRUE`.
#' @param table An Arrow `Table`.
#' @return A list with `table`, `model_version`, `request_type`, `gene_order`.
#' @noRd
arrow_raw_result <- function(table) {
  metadata <- arrow_schema_metadata(table)
  list(
    table = table,
    model_version = metadata$model_version,
    request_type = metadata$request_type,
    gene_order = parse_gene_order(metadata$gene_order)
  )
}

#' @title Run a Synchronous Self-Hosted Prediction (Internal)
#' @description Performs a single synchronous POST to a self-hosted model
#' container's predict endpoint (`/api/models/<model_id>/predict`), requesting an
#' Apache Arrow IPC stream, and converts the response into output data frames. Unlike
#' the production path this does NOT require an API key; a key is only attached
#' when one is configured.
#' @param query The query list to send as the JSON request body.
#' @param model_id Character string specifying the model ID.
#' @param api_base_url The base URL for the self-hosted container.
#' @param raw_response Logical; if TRUE, return the parsed Arrow `Table` and
#'        schema metadata without applying any transformer.
#' @return A named list of data.frames, or the raw Arrow result when
#'         `raw_response = TRUE`.
#' @importFrom httr POST content status_code timeout
#' @importFrom jsonlite toJSON
#' @noRd
predict_query_self_hosted <- function(query, model_id, api_base_url, raw_response = FALSE) {
  require_arrow()

  url <- paste0(api_base_url, "/api/models/", model_id, "/predict")
  query_json <- toJSON(query, auto_unbox = TRUE)

  response <- tryCatch(
    POST(
      url = url,
      build_arrow_post_headers(),
      body = query_json,
      encode = "json",
      timeout(SELF_HOSTED_TIMEOUT)
    ),
    error = function(e) {
      stop(paste0("Self-hosted predict request failed due to a network issue: ", e$message))
    }
  )

  if (status_code(response) >= 400) {
    stop(paste0(
      "Self-hosted predict request failed with status ",
      status_code(response), ": ", content(response, "text")
    ))
  }

  raw_bytes <- content(response, as = "raw")
  table <- read_arrow_stream(raw_bytes)

  if (isTRUE(raw_response)) {
    return(arrow_raw_result(table))
  }

  transform_arrow_table(table)
}

#' @title Set Synthesize Bio API Token
#' @description Securely prompts for and stores the Synthesize Bio API token in the
#' environment. This function uses getPass to securely handle the token input without
#' displaying it in the console. The token is stored in the SYNTHESIZE_API_KEY
#' environment variable for the current R session.
#'
#' @param use_keyring Logical, whether to also store the token securely in the system
#'        keyring for future sessions. Defaults to FALSE.
#' @param token Character, optional. If provided, uses this token instead of prompting.
#'        This parameter should only be used in non-interactive scripts.
#'
#' @return Invisibly returns TRUE if successful.
#' @importFrom getPass getPass
#' @importFrom utils browseURL
#' @examples
#' # Interactive prompt for token
#' \dontrun{
#' set_synthesize_token()
#'
#' # Provide token directly (less secure, not recommended for interactive use)
#' set_synthesize_token(token = "your-token-here")
#'
#' # Store in system keyring for future sessions
#' set_synthesize_token(use_keyring = TRUE)
#' }
#' @export
set_synthesize_token <- function(use_keyring = FALSE, token = NULL) {
  if (is.null(token)) {
    message("Create an account at https://app.synthesize.bio/ then go to your account page.")
    message("Click \"Create API key\" on the API keys tab.")

    browseURL("https://app.synthesize.bio/account/api-keys")
    token <- getPass::getPass(msg = paste(
      "Create an account at https://app.synthesize.bio/ then go to your account page.",
      "Click create token then copy it.",
      "Paste token here and press enter: "
    ))
  }

  # Store in environment
  Sys.setenv(SYNTHESIZE_API_KEY = token)

  # Optionally store in keyring if requested and available
  if (use_keyring) {
    if (requireNamespace("keyring", quietly = TRUE)) {
      tryCatch(
        {
          keyring::key_set_with_value(
            service = "rsynthbio",
            username = "api_token",
            password = token
          )
          message("API token stored in system keyring.")
        },
        error = function(e) {
          warning("Failed to store token in keyring: ", e$message)
        }
      )
    } else {
      warning("Package 'keyring' is not installed. Token not stored in keyring.")
      message("To store token in keyring, install with: install.packages('keyring')")
    }
  }

  message("API token set for current session.")
  invisible(TRUE)
}

#' @title Load Synthesize Bio API Token from Keyring
#' @description Loads the previously stored Synthesize Bio API token from the system
#' keyring and sets it in the environment for the current session.
#'
#' @return Invisibly returns TRUE if successful, FALSE if token not found in keyring.
#' @examples
#' \dontrun{
#' # Load token from keyring
#' load_synthesize_token_from_keyring()
#' }
#' @export
load_synthesize_token_from_keyring <- function() {
  if (!requireNamespace("keyring", quietly = TRUE)) {
    warning("Package 'keyring' is not installed. Cannot load token from keyring.")
    message("To use this feature, install with: install.packages('keyring')")
    return(invisible(FALSE))
  }

  tryCatch(
    {
      token <- keyring::key_get(service = "rsynthbio", username = "api_token")
      Sys.setenv(SYNTHESIZE_API_KEY = token)
      message("API token loaded from keyring and set for current session.")
      invisible(TRUE)
    },
    error = function(e) {
      warning("Failed to load token from keyring: ", e$message)
      invisible(FALSE)
    }
  )
}

#' @title Clear Synthesize Bio API Token
#' @description Clears the Synthesize Bio API token from the environment for the
#' current R session. This is useful for security purposes when you've finished
#' working with the API or when switching between different accounts.
#'
#' @param remove_from_keyring Logical, whether to also remove the token from the
#'        system keyring if it's stored there. Defaults to FALSE.
#'
#' @return Invisibly returns TRUE.
#' @examples
#' \dontrun{
#' # Clear token from current session only
#' clear_synthesize_token()
#'
#' # Clear token from both session and keyring
#' clear_synthesize_token(remove_from_keyring = TRUE)
#' }
#' @export
clear_synthesize_token <- function(remove_from_keyring = FALSE) {
  # Unset from environment
  if (Sys.getenv("SYNTHESIZE_API_KEY") != "") {
    Sys.unsetenv("SYNTHESIZE_API_KEY")
    message("API token cleared from current session.")
  } else {
    message("No API token was set in the current session.")
  }

  # Optionally remove from keyring
  if (remove_from_keyring) {
    if (requireNamespace("keyring", quietly = TRUE)) {
      tryCatch(
        {
          if (keyring::key_list(service = "rsynthbio")[1, "username"] == "api_token") {
            keyring::key_delete(service = "rsynthbio", username = "api_token")
            message("API token removed from system keyring.")
          } else {
            message("No API token was found in the keyring.")
          }
        },
        error = function(e) {
          # This might occur if no token exists or other keyring issues
          message("No API token was found in the keyring or could not access keyring.")
        }
      )
    } else {
      warning("Package 'keyring' is not installed. Cannot remove token from keyring.")
      message("To use this feature, install with: install.packages('keyring')")
    }
  }

  invisible(TRUE)
}

#' @title Check if Synthesize Bio API Token is Set
#' @description Checks whether a Synthesize Bio API token is currently set in the
#' environment. Useful for conditional code that requires an API token.
#'
#' @return Logical, TRUE if token is set, FALSE otherwise.
#' @examples  \dontrun{
#' # Check if token is set
#' if (!has_synthesize_token()) {
#'   # Prompt for token if not set
#'   set_synthesize_token()
#' }
#' }
#' @export
has_synthesize_token <- function() {
  Sys.getenv("SYNTHESIZE_API_KEY") != ""
}

library(mockery)

# Mock for environment variables
mock_env <- function(expr, envvar, value) {
  old <- Sys.getenv(envvar, unset = NA)
  Sys.setenv(!!envvar := value)
  on.exit(if (is.na(old)) Sys.unsetenv(envvar) else Sys.setenv(!!envvar := old))
  force(expr)
}

# Mock for getPass
mock_getpass <- function() {
  "mock-token-12345"
}

# Setup mock for keyring package
mock_keyring <- list(
  stored_keys = list(),
  key_set_with_value = function(service, username, password) {
    mock_keyring$stored_keys[[paste0(service, ":", username)]] <- password
    invisible(TRUE)
  },
  key_get = function(service, username) {
    key <- mock_keyring$stored_keys[[paste0(service, ":", username)]]
    if (is.null(key)) stop("Key not found")
    return(key)
  },
  key_delete = function(service, username) {
    mock_keyring$stored_keys[[paste0(service, ":", username)]] <- NULL
    invisible(TRUE)
  },
  key_list = function(service) {
    keys <- names(mock_keyring$stored_keys)
    keys <- keys[grepl(paste0("^", service, ":"), keys)]
    if (length(keys) == 0) {
      return(data.frame(service = character(0), username = character(0)))
    }
    usernames <- sub(paste0("^", service, ":"), "", keys)
    return(data.frame(service = rep(service, length(usernames)),
                      username = usernames,
                      stringsAsFactors = FALSE))
  }
)

test_that("set_synthesize_token stores token in environment", {
  # Save original to restore after test
  orig_env <- Sys.getenv("SYNTHESIZE_API_KEY", unset = NA)
  on.exit(if (is.na(orig_env)) Sys.unsetenv("SYNTHESIZE_API_KEY") else Sys.setenv(SYNTHESIZE_API_KEY = orig_env))

  # Clear environment before test
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Test with direct token
  expect_message(set_synthesize_token(token = "test-token"), "API token set for current session")
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "test-token")
})

test_that("set_synthesize_token prompts for token when not provided", {

  # Save original to restore after test
  orig_env <- Sys.getenv("SYNTHESIZE_API_KEY", unset = NA)
  on.exit(if (is.na(orig_env)) Sys.unsetenv("SYNTHESIZE_API_KEY") else Sys.setenv(SYNTHESIZE_API_KEY = orig_env))

  # Clear environment before test
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Mock getPass function to return a predefined token
  m <- mock(mock_getpass())
  stub(set_synthesize_token, "getPass::getPass", m)
  stub(set_synthesize_token, "browseURL", function(...) NULL)

  # Call the function and verify it used our mock to get the token
  expect_message(set_synthesize_token(), "API token set for current session")
  expect_called(m, 1)
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "mock-token-12345")
})

test_that("set_synthesize_token stores token in keyring when requested", {

  # Save original to restore after test
  orig_env <- Sys.getenv("SYNTHESIZE_API_KEY", unset = NA)
  on.exit(if (is.na(orig_env)) Sys.unsetenv("SYNTHESIZE_API_KEY") else Sys.setenv(SYNTHESIZE_API_KEY = orig_env))

  # Clear environment before test
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Mock requireNamespace to always return TRUE for keyring
  m_req <- mock(TRUE)
  stub(set_synthesize_token, "requireNamespace", m_req)

  # Mock keyring functions
  m_set <- mock(TRUE)
  stub(set_synthesize_token, "keyring::key_set_with_value", m_set)

  # Call the function with keyring storage enabled
  expect_message(
    set_synthesize_token(use_keyring = TRUE, token = "keyring-test-token"),
    "API token stored in system keyring"
  )

  # Verify keyring function was called
  expect_called(m_req, 1)
  expect_called(m_set, 1)
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "keyring-test-token")
})

test_that("set_synthesize_token handles missing keyring package", {

  # Save original to restore after test
  orig_env <- Sys.getenv("SYNTHESIZE_API_KEY", unset = NA)
  on.exit(if (is.na(orig_env)) Sys.unsetenv("SYNTHESIZE_API_KEY") else Sys.setenv(SYNTHESIZE_API_KEY = orig_env))

  # Clear environment before test
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Mock requireNamespace to return FALSE for keyring
  m_req <- mock(FALSE)
  stub(set_synthesize_token, "requireNamespace", m_req)

  # Call the function with keyring storage enabled, but keyring not available
  expect_warning(
    set_synthesize_token(use_keyring = TRUE, token = "unavailable-keyring-token"),
    "Package 'keyring' is not installed"
  )

  expect_called(m_req, 1)
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "unavailable-keyring-token")
})

test_that("load_synthesize_token_from_keyring loads token from keyring", {

  # Save original to restore after test
  orig_env <- Sys.getenv("SYNTHESIZE_API_KEY", unset = NA)
  on.exit(if (is.na(orig_env)) Sys.unsetenv("SYNTHESIZE_API_KEY") else Sys.setenv(SYNTHESIZE_API_KEY = orig_env))

  # Clear environment before test
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Mock requireNamespace to return TRUE for keyring
  m_req <- mock(TRUE)
  stub(load_synthesize_token_from_keyring, "requireNamespace", m_req)

  # Mock keyring::key_get to return a token
  m_get <- mock("keyring-stored-token")
  stub(load_synthesize_token_from_keyring, "keyring::key_get", m_get)

  # Call the function
  expect_message(
    result <- load_synthesize_token_from_keyring(),
    "API token loaded from keyring"
  )

  # Verify keyring function was called and token was set
  expect_true(result)
  expect_called(m_req, 1)
  expect_called(m_get, 1)
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "keyring-stored-token")
})

test_that("load_synthesize_token_from_keyring handles missing keyring package", {

  # Save original to restore after test
  orig_env <- Sys.getenv("SYNTHESIZE_API_KEY", unset = NA)
  on.exit(if (is.na(orig_env)) Sys.unsetenv("SYNTHESIZE_API_KEY") else Sys.setenv(SYNTHESIZE_API_KEY = orig_env))

  # Clear environment before test
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Mock requireNamespace to return FALSE for keyring
  m_req <- mock(FALSE)
  stub(load_synthesize_token_from_keyring, "requireNamespace", m_req)

  # Call the function
  expect_warning(
    result <- load_synthesize_token_from_keyring(),
    "Package 'keyring' is not installed"
  )

  # Verify function returns FALSE and environment is unchanged
  expect_false(result)
  expect_called(m_req, 1)
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "")
})

test_that("load_synthesize_token_from_keyring handles keyring errors", {

  # Save original to restore after test
  orig_env <- Sys.getenv("SYNTHESIZE_API_KEY", unset = NA)
  on.exit(if (is.na(orig_env)) Sys.unsetenv("SYNTHESIZE_API_KEY") else Sys.setenv(SYNTHESIZE_API_KEY = orig_env))

  # Clear environment before test
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Mock requireNamespace to return TRUE for keyring
  m_req <- mock(TRUE)
  stub(load_synthesize_token_from_keyring, "requireNamespace", m_req)

  # Verify function returns FALSE and environment is unchanged
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "")
})

test_that("clear_synthesize_token clears token from environment", {
  # Set a token
  Sys.setenv(SYNTHESIZE_API_KEY = "token-to-clear")

  # Clear it
  expect_message(clear_synthesize_token(), "API token cleared from current session")

  # Verify it's gone
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "")
})

test_that("clear_synthesize_token handles already empty environment", {
  # Make sure no token is set
  Sys.unsetenv("SYNTHESIZE_API_KEY")

  # Try to clear it
  expect_message(clear_synthesize_token(), "No API token was set in the current session")

  # Verify still empty
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "")
})

test_that("clear_synthesize_token removes token from keyring when requested", {

  # Set a token
  Sys.setenv(SYNTHESIZE_API_KEY = "token-to-clear-from-keyring")

  # Mock requireNamespace to return TRUE for keyring
  m_req <- mock(TRUE)
  stub(clear_synthesize_token, "requireNamespace", m_req)

  # Mock keyring functions
  m_list <- mock(data.frame(service = "rsynthbio", username = "api_token", stringsAsFactors = FALSE))
  stub(clear_synthesize_token, "keyring::key_list", m_list)

  m_delete <- mock(TRUE)
  stub(clear_synthesize_token, "keyring::key_delete", m_delete)

  # Call the function with keyring removal enabled
  expect_message(
    clear_synthesize_token(remove_from_keyring = TRUE),
    "API token removed from system keyring"
  )

  # Verify functions were called
  expect_called(m_req, 1)
  expect_called(m_list, 1)
  expect_called(m_delete, 1)
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "")
})

test_that("clear_synthesize_token handles missing keyring package", {

  # Set a token
  Sys.setenv(SYNTHESIZE_API_KEY = "token-to-clear-no-keyring")

  # Mock requireNamespace to return FALSE for keyring
  m_req <- mock(FALSE)
  stub(clear_synthesize_token, "requireNamespace", m_req)

  # Call the function with keyring removal enabled, but keyring not available
  expect_warning(
    clear_synthesize_token(remove_from_keyring = TRUE),
    "Package 'keyring' is not installed"
  )

  # Verify functions were called appropriately
  expect_called(m_req, 1)
  expect_equal(Sys.getenv("SYNTHESIZE_API_KEY"), "")
})

test_that("has_synthesize_token detects token presence correctly", {
  # Test when token is not set
  Sys.unsetenv("SYNTHESIZE_API_KEY")
  expect_false(has_synthesize_token())

  # Test when token is set
  Sys.setenv(SYNTHESIZE_API_KEY = "test-token-existence")
  expect_true(has_synthesize_token())

  # Clean up
  Sys.unsetenv("SYNTHESIZE_API_KEY")
})

---
title: "Available Metadata"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Available Metadata}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Overview

Every model on Synthesize Bio accepts specific metadata fields as inputs—things like cell type, tissue, disease state, and perturbation details. The valid values for each field (e.g. ontology IDs, cell line names) are defined by the model's vocabulary.

You can browse and download the vocabulary for any model you have access to at:

**[app.synthesize.bio/docs/vocab](https://app.synthesize.bio/docs/vocab)**

Select a model from the dropdown to see all available fields, and use the download links to get the full list of valid values for each field as JSON.

---
title: "Baseline Models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Baseline Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)
```

## Overview

Baseline models generate synthetic gene expression data from metadata alone. You describe the biological conditions—tissue type, disease state, perturbations, cell type, etc.—and the model generates realistic expression profiles matching those conditions.

This is the most common use case: generating synthetic data for conditions where real data may be scarce or unavailable.

## Available Models

- **`gem-1-bulk`**: Bulk RNA-seq baseline model
- **`gem-1-sc`**: Single-cell RNA-seq baseline model

```{r}
library(rsynthbio)
```

## Creating a Query

The structure of the query required by the API is specific to each model. Use `get_example_query()` to get a correctly structured example for your chosen model.

```{r query-example}
# Get the example query structure for a specific model
example_query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Inspect the query structure
str(example_query)
```

The query consists of:

1. **`sampling_strategy`**: The prediction mode that controls how expression data is generated:
   - **"sample generation"**: Generates realistic-looking synthetic data with measurement error (bulk only)
   - **"mean estimation"**: Provides stable mean estimates of expression levels (bulk and single-cell)
2. **`inputs`**: A list of biological conditions to generate data for

Each input contains `metadata` (describing the biological sample) and `num_samples` (how many samples to generate).

## Making a Prediction

Once your query is ready, send it to the API to generate gene expression data:

```{r predict, eval=FALSE}
# Create a query for the bulk model
query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Submit and get results
result <- predict_query(query, model_id = "gem-1-bulk")
```

The result is a list containing two data frames: `metadata` and `expression`.

### Single-Cell Example

```{r sc-example, eval=FALSE}
# Create a query for the single-cell model
sc_query <- get_example_query(model_id = "gem-1-sc")$example_query

# Submit and get results
sc_result <- predict_query(sc_query, model_id = "gem-1-sc")
```

> **Note:** Single-cell models only support `"mean estimation"` mode.

## Query Parameters

In addition to metadata, queries support several optional parameters that control the generation process.

### sampling_strategy (character, required)

Controls the type of prediction the model generates. This parameter is required in all queries.

Available modes:

- **"sample generation"**: The model generates realistic-looking synthetic data that captures measurement error. This mode is useful when you want data that mimics real experimental measurements. **(Bulk only)**

- **"mean estimation"**: The model creates a distribution capturing biological heterogeneity consistent with the supplied metadata, then returns the mean of that distribution. This mode is useful when you want a stable estimate of expected expression levels. **(Bulk and single-cell)**

```{r mode-examples, eval=FALSE}
# Bulk query with sample generation
bulk_query <- get_example_query(model_id = "gem-1-bulk")$example_query
bulk_query$sampling_strategy <- "sample generation"

# Bulk query with mean estimation
bulk_query_mean <- get_example_query(model_id = "gem-1-bulk")$example_query
bulk_query_mean$sampling_strategy <- "mean estimation"

# Single-cell query (must use mean estimation)
sc_query <- get_example_query(model_id = "gem-1-sc")$example_query
sc_query$sampling_strategy <- "mean estimation" # Required for single-cell
```

### total_count (integer, optional)

Library size used when converting predicted log CPM back to raw counts. Higher values scale counts up proportionally.

- Default: 10,000,000 for bulk; 10,000 for single-cell

```{r total-count-example, eval=FALSE}
# Create a query and add custom total_count
query <- get_example_query(model_id = "gem-1-bulk")$example_query
query$total_count <- 5000000
```

### deterministic_latents (logical, optional)

If `TRUE`, the model uses the mean of each latent distribution (`p(z|metadata)`) instead of sampling. This removes randomness from latent sampling and produces deterministic outputs for the same inputs.

- Default: `FALSE` (sampling is enabled)

```{r deterministic-example, eval=FALSE}
# Create a query and enable deterministic latents
query <- get_example_query(model_id = "gem-1-bulk")$example_query
query$deterministic_latents <- TRUE
```

### seed (integer, optional)

Random seed for reproducibility when using stochastic sampling.

```{r seed-example, eval=FALSE}
# Create a query with a specific seed
query <- get_example_query(model_id = "gem-1-bulk")$example_query
query$seed <- 42
```

### Combining Parameters

You can combine multiple parameters in a single query:

```{r combined-params, eval=FALSE}
# Create a query and add multiple parameters
query <- get_example_query(model_id = "gem-1-bulk")$example_query
query$total_count <- 8000000
query$deterministic_latents <- TRUE
query$sampling_strategy <- "mean estimation"

results <- predict_query(query, model_id = "gem-1-bulk")
```

## Valid Metadata Keys

The input metadata is a list of lists. Here is the full list of valid metadata keys:

### Biological

- `age_years`
- `cell_line_ontology_id`
- `cell_type_ontology_id`
- `developmental_stage`
- `disease_ontology_id`
- `ethnicity`
- `genotype`
- `race`
- `sample_type` ("cell line", "organoid", "other", "primary cells", "primary tissue", "xenograft")
- `sex` ("male", "female")
- `tissue_ontology_id`

### Perturbational

- `perturbation_dose` (number and unit separated by a space, e.g., "10 um")
- `perturbation_ontology_id`
- `perturbation_time` (number and unit separated by a space, e.g., "24 hours")
- `perturbation_type` ("coculture", "compound", "control", "crispr", "genetic", "infection", "other", "overexpression", "peptide or biologic", "shrna", "sirna")

### Technical

- `study` (Bioproject ID)
- `library_selection` (e.g., "cDNA", "polyA", "Oligo-dT" - see https://ena-docs.readthedocs.io/en/latest/submit/reads/webin-cli.html#permitted-values-for-library-selection)
- `library_layout` ("PAIRED", "SINGLE")
- `platform` ("illumina")

## Valid Metadata Values

The following are the valid values or expected formats for selected metadata keys:

| Metadata Field             | Requirement / Example                                                                                                                                                                                                 |
|----------------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| `cell_line_ontology_id`    | Requires a [Cellosaurus ID](https://www.cellosaurus.org/).                                                                                                                                                           |
| `cell_type_ontology_id`    | Requires a [CL ID](https://www.ebi.ac.uk/ols4/ontologies/cl).                                                                                                                                                         |
| `disease_ontology_id`      | Requires a [MONDO ID](https://www.ebi.ac.uk/ols4/ontologies/mondo).                                                                                                                                                   |
| `perturbation_ontology_id` | Must be a valid Ensembl gene ID (e.g., `ENSG00000156127`), [ChEBI ID](https://www.ebi.ac.uk/chebi/) (e.g., `CHEBI:16681`), [ChEMBL ID](https://www.ebi.ac.uk/chembl/) (e.g., `CHEMBL1234567`), or [NCBI Taxonomy ID](https://www.ncbi.nlm.nih.gov/taxonomy) (e.g., `9606`). |
| `tissue_ontology_id`       | Requires a [UBERON ID](https://www.ebi.ac.uk/ols4/ontologies/uberon).                                                                                                                                                 |

We highly recommend using the [EMBL-EBI Ontology Lookup Service](https://www.ebi.ac.uk/ols4/) to find valid IDs for your metadata.

Models have a limited acceptable range of metadata input values. If you provide a value that is not in the acceptable range, the API will return an error.

## Modifying Query Inputs

You can customize the query inputs to fit your specific research needs:

```{r modify-query, eval=FALSE}
# Get a base query
query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Adjust number of samples for the first input
query$inputs[[1]]$num_samples <- 10

# Add a new condition
query$inputs[[3]] <- list(
  metadata = list(
    sex = "male",
    sample_type = "primary tissue",
    tissue_ontology_id = "UBERON:0002371"
  ),
  num_samples = 5
)
```

## Working with Results

```{r analyze, eval=FALSE}
# Access metadata and expression matrices
metadata <- result$metadata
expression <- result$expression

# Check dimensions
dim(expression)

# View metadata sample
head(metadata)
```

You may want to process the data in chunks or save it for later use:

```{r large-data, eval=FALSE}
# Save results to RDS file
saveRDS(result, "synthesize_results.rds")

# Load previously saved results
result <- readRDS("synthesize_results.rds")

# Export as CSV
write.csv(result$expression, "expression_matrix.csv")
write.csv(result$metadata, "sample_metadata.csv")
```

---
title: "Self-Hosted Models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Self-Hosted Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since these calls require a running container
)
```

Partners who run Synthesize Bio models inside their own environment (for
example, on a GPU host in their own cloud account) can use the same `rsynthbio`
client against a self-hosted model container instead of the hosted API at
`app.synthesize.bio`.

> **Self-hosted deployment is a model deployment option available within a
> Synthesize Bio partnership.** To learn more or request access, contact
> [partnerships@synthesize.bio](mailto:partnerships@synthesize.bio).

## How it differs from the hosted path

The hosted path is asynchronous: it starts a query, polls for completion, and
downloads results. A self-hosted container instead returns predictions
**synchronously** as an [Apache Arrow](https://arrow.apache.org/) IPC stream.
`rsynthbio` decodes that stream into exactly the same data frames you get from
the hosted path (`expression`, `metadata`, and `latents`), so downstream code
does not change.

Key differences:

- **No polling and no download URL** — a single request returns the data.
- **Requires the optional `arrow` package** — install it with
  `install.packages("arrow")`.
- **No API key required** — a key is only sent when `SYNTHESIZE_API_KEY` is set
  (use this if your container runs with authentication enabled).

## Installation

```{r install, eval=FALSE}
install.packages("rsynthbio")
install.packages("arrow") # required for self-hosted mode
```

## Enabling self-hosted mode

Set `self_hosted = TRUE` on a call, or enable it for the whole session with the
`SYNTHESIZE_SELF_HOSTED` environment variable (truthy values: `1`, `true`,
`yes`, `on`).

```{r enable, eval=FALSE}
library(rsynthbio)

Sys.setenv(SYNTHESIZE_SELF_HOSTED = "1")
```

## Pointing each model at its container

Self-hosted deployments typically run one container per model. Set a per-model
base URL once and you never have to pass `api_base_url` on individual calls.
The variable name is `SYNTHESIZE_API_BASE_URL__<MODEL>`, where `<MODEL>` is the
upper-cased model id with non-alphanumeric characters replaced by underscores
(for example, `gem-1-bulk` becomes `SYNTHESIZE_API_BASE_URL__GEM_1_BULK`).

```{r per-model, eval=FALSE}
Sys.setenv(
  SYNTHESIZE_API_BASE_URL__GEM_1_BULK = "https://gem-1-bulk.internal.example",
  SYNTHESIZE_API_BASE_URL__GEM_1_SC   = "https://gem-1-sc.internal.example"
)

query <- get_example_query("gem-1-bulk", self_hosted = TRUE)$example_query
result <- predict_query(query, model_id = "gem-1-bulk", self_hosted = TRUE)

expression <- result$expression
metadata <- result$metadata
```

Variant slugs backed by the same container (for example,
`gem-1-bulk_reference-conditioning` and `gem-1-bulk_predict-metadata`) resolve
to the same per-model variable as their base model.

### Resolution precedence

When `api_base_url` is `NULL`, the base URL is resolved in this order:

1. An explicit `api_base_url` argument passed to the call.
2. The per-model variable `SYNTHESIZE_API_BASE_URL__<MODEL>`.
3. The global `SYNTHESIZE_API_BASE_URL`.
4. The production default (`https://app.synthesize.bio`).

You can always override the environment by passing `api_base_url` directly:

```{r explicit-url, eval=FALSE}
result <- predict_query(
  query,
  model_id = "gem-1-bulk",
  api_base_url = "https://gem-1-bulk.internal.example",
  self_hosted = TRUE
)
```

## Authentication (optional)

Self-hosted containers may run without authentication. If yours requires a key,
set `SYNTHESIZE_API_KEY` and the client attaches it as a bearer token:

```{r auth, eval=FALSE}
Sys.setenv(SYNTHESIZE_API_KEY = "your-container-api-key")
```

## Raw responses

Pass `raw_response = TRUE` to receive the parsed Arrow `Table` and its schema
metadata (including `model_version`, `request_type`, and `gene_order`) instead
of the transformed data frames:

```{r raw, eval=FALSE}
raw <- predict_query(
  query,
  model_id = "gem-1-bulk",
  self_hosted = TRUE,
  raw_response = TRUE
)

raw$table # an arrow::Table
raw$model_version
```

## Session info

```{r session-info}
sessionInfo()
```

---
title: "Reference Conditioning"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reference Conditioning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)
```

## Overview

Reference conditioning models generate expression data **conditioned on a real reference sample**. This allows you to "anchor" to an existing expression profile while applying perturbations or modifications.

This is useful when you want to:

- Simulate the effect of a perturbation on a specific sample
- Generate expression profiles that preserve the biological and technical characteristics of a reference
- Create synthetic "treated vs. control" pairs

## Available Models

- **`gem-1-bulk_reference-conditioning`**: Bulk RNA-seq reference conditioning model
- **`gem-1-sc_reference-conditioning`**: Single-cell RNA-seq reference conditioning model

> **Note:** These endpoints may require 1-2 minutes of startup time if they have been scaled down. Plan accordingly for interactive use.

```{r}
library(rsynthbio)
```

## How It Works

Reference conditioning encodes the biological and technical characteristics from a real expression sample, then generates new expression data that:

1. Preserves the biological/technical latent space of the reference
2. Applies any perturbation metadata you specify
3. Returns synthetic expression that reflects the perturbation effect on that specific sample

## Creating a Query

Reference conditioning queries require different inputs than baseline models:

```{r query-example, eval=FALSE}
# Get the example query structure
example_query <- get_example_query(model_id = "gem-1-bulk_reference-conditioning")$example_query

# Inspect the query structure
str(example_query)
```

The query structure includes:

1. **`inputs`**: A list where each input contains:
   - **`counts`**: The reference expression counts (a numeric vector)
   - **`metadata`**: Perturbation-only metadata (see below)
   - **`num_samples`**: How many samples to generate

2. **`conditioning`**: Which latent spaces to condition on (typically `["biological", "technical"]`)

3. **`sampling_strategy`**: `"mean estimation"` or `"sample generation"`

### Perturbation-Only Metadata

Unlike baseline models, reference conditioning queries only accept perturbation metadata fields:

- `perturbation_ontology_id`
- `perturbation_type`
- `perturbation_time`
- `perturbation_dose`

All other biological and technical metadata is inferred from the reference expression.

## Example: Simulating a Drug Treatment

Here's a complete example simulating a drug treatment effect on a reference sample:

```{r drug-treatment, eval=FALSE}
# Start with example query structure
query <- get_example_query(model_id = "gem-1-bulk_reference-conditioning")$example_query

# Replace with your actual reference counts
# The counts vector must match the model's expected gene order and length
query$inputs[[1]]$counts <- your_reference_counts

# Specify the perturbation
query$inputs[[1]]$metadata <- list(
  perturbation_ontology_id = "CHEMBL25", # Aspirin (ChEMBL ID)
  perturbation_type = "compound",
  perturbation_time = "24 hours",
  perturbation_dose = "10 um"
)

query$inputs[[1]]$num_samples <- 3

# Set the sampling strategy
query$sampling_strategy <- "mean estimation"

# Submit the query
result <- predict_query(query, model_id = "gem-1-bulk_reference-conditioning")
```

## Example: CRISPR Knockout Simulation

Simulate the effect of knocking out a specific gene:

```{r crispr-example, eval=FALSE}
query <- get_example_query(model_id = "gem-1-bulk_reference-conditioning")$example_query

# Your reference sample counts
query$inputs[[1]]$counts <- control_sample_counts

# CRISPR knockout of TP53
query$inputs[[1]]$metadata <- list(
  perturbation_ontology_id = "ENSG00000141510", # TP53 Ensembl ID
  perturbation_type = "crispr"
)

query$inputs[[1]]$num_samples <- 5

result <- predict_query(query, model_id = "gem-1-bulk_reference-conditioning")
```

## Query Parameters

### conditioning (list, optional)

Controls which latent spaces are conditioned on the reference. Default is `["biological", "technical"]`.

When both are conditioned, the model preserves both biological identity and technical characteristics from the reference sample.

### sampling_strategy (character, required)

Controls the type of prediction:

- **"sample generation"**: Generates realistic-looking synthetic data with measurement error. **(Bulk only)**
- **"mean estimation"**: Provides stable mean estimates. **(Bulk and single-cell)**

```{r mode-example, eval=FALSE}
query$sampling_strategy <- "mean estimation"
```

### fixed_total_count (logical, optional)

Controls whether to preserve the reference's library size:

- **`FALSE`** (default): The output's total count is taken from the reference expression (sum of its counts). Use this when you want the synthetic sample to preserve the reference's library size.
- **`TRUE`**: Forces the model to use the `total_count` parameter value (or default) instead of the reference's library size.

```{r fixed-total-count, eval=FALSE}
# Preserve reference library size (default)
query$fixed_total_count <- FALSE

# Or force a specific library size
query$fixed_total_count <- TRUE
query$total_count <- 10000000
```

### total_count (integer, optional)

Library size used when converting predicted log CPM back to raw counts. Only effective when `fixed_total_count = TRUE`.

- Default: 10,000,000 for bulk; 10,000 for single-cell

### deterministic_latents (logical, optional)

If `TRUE`, the model uses the mean of each latent distribution (`p(z|metadata)` for perturbation, `q(z|x)` for conditioned components) instead of sampling. This produces deterministic, reproducible outputs.

- Default: `FALSE`

```{r deterministic-example, eval=FALSE}
query$deterministic_latents <- TRUE
```

### seed (integer, optional)

Random seed for reproducibility.

```{r seed-example, eval=FALSE}
query$seed <- 42
```

## Valid Perturbation Metadata

| Field                      | Description / Format                                                                                                                                                                  |
|----------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| `perturbation_ontology_id` | Ensembl gene ID (e.g., `ENSG00000141510`), [ChEBI ID](https://www.ebi.ac.uk/chebi/), [ChEMBL ID](https://www.ebi.ac.uk/chembl/), or [NCBI Taxonomy ID](https://www.ncbi.nlm.nih.gov/taxonomy) |
| `perturbation_type`        | One of: "coculture", "compound", "control", "crispr", "genetic", "infection", "other", "overexpression", "peptide or biologic", "shrna", "sirna"                                      |
| `perturbation_time`        | Time since perturbation, as a number and unit separated by a space (e.g., "24 hours", "48 hours")                                                                                      |
| `perturbation_dose`        | Dose of perturbation, as a number and unit separated by a space (e.g., "10 um", "1 mg/kg")                                                                                             |

## Working with Results

The result structure is similar to baseline models:

```{r results, eval=FALSE}
# Access metadata and expression matrices
metadata <- result$metadata
expression <- result$expression

# Compare to your reference
dim(expression)
head(metadata)
```

### Differential Expression

When conditioning on both biological and technical latents, you can directly compare the generated expression to your reference to identify perturbation effects:
```{r de-analysis, eval=FALSE}
# Your reference (input) counts
reference_cpm <- your_reference_counts / sum(your_reference_counts) * 1e6

# Generated (perturbed) counts
generated_cpm <- expression[1, ] / sum(expression[1, ]) * 1e6

# Log fold change
log2fc <- log2(generated_cpm + 1) - log2(reference_cpm + 1)

# Identify top changed genes
head(sort(log2fc, decreasing = TRUE), 20)
```

## Important Notes

### Counts Vector Length

The reference counts vector must match the model's expected number of genes. If the length doesn't match, the API will return a validation error.

Use `get_example_query()` to see the expected structure and ensure your counts vector has the correct length.

### Gene Order

Ensure your reference counts are in the same gene order expected by the model. The response includes a `gene_order` field that specifies the expected order.

---
title: "Getting Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)
```


`rsynthbio` is an R package that provides a convenient interface to the [Synthesize Bio](https://www.synthesize.bio/) API, allowing users to generate realistic gene expression data based on specified biological conditions.
This package enables researchers to easily access AI-generated transcriptomic data for various modalities including bulk RNA-seq and single-cell RNA-seq.

Alternatively, you can AI generate datasets from our [web platform](https://app.synthesize.bio/datasets/).

## How to install

You can install `rsynthbio` from CRAN:

```{r installation, eval=FALSE}
install.packages("rsynthbio")
```

If you want the development version, you can install using the `remotes` package to install from GitHub:

```{r github-installation, eval=FALSE}
if (!("remotes" %in% installed.packages())) {
  install.packages("remotes")
}
remotes::install_github("synthesizebio/rsynthbio")
```

Once installed, load the package:

```{r}
library(rsynthbio)
```

## Authentication

Before using the Synthesize Bio API, you need to set up your API token. The package provides a secure way to handle authentication:

```{r auth-secure, eval=FALSE}
# Securely prompt for and store your API token
# The token will not be visible in the console
set_synthesize_token()

# You can also store the token in your system keyring for persistence
# across R sessions (requires the 'keyring' package)
set_synthesize_token(use_keyring = TRUE)
```

Loading your API key for a session.

```{r, eval=FALSE}
# In future sessions, load the stored token
load_synthesize_token_from_keyring()

# Check if a token is already set
has_synthesize_token()
```

You can manually set the token, but don't commit it to version control!
```{r, eval=FALSE}
set_synthesize_token(token = "your-token-here")
```

You can obtain an API token by registering at [Synthesize Bio](https://app.synthesize.bio).

## Available Model Types

Synthesize Bio provides several types of models for different use cases:

### Baseline Models

Generate synthetic gene expression data from metadata alone. You describe the biological conditions (tissue type, disease state, perturbations, etc.) and the model generates realistic expression profiles.

- **`gem-1-bulk`**: Bulk RNA-seq baseline model
- **`gem-1-sc`**: Single-cell RNA-seq baseline model

See the [Baseline Models](baseline.html) vignette for detailed usage.

### Reference Conditioning Models

Generate expression data conditioned on a real reference sample. This allows you to "anchor" to an existing expression profile while applying perturbations or modifications.

- **`gem-1-bulk_reference-conditioning`**: Bulk RNA-seq reference conditioning model
- **`gem-1-sc_reference-conditioning`**: Single-cell RNA-seq reference conditioning model

See the [Reference Conditioning](reference-conditioning.html) vignette for detailed usage.

### Metadata Prediction Models

Infer metadata from observed expression data. Given a gene expression profile, predict the likely biological characteristics (cell type, tissue, disease state, etc.).

- **`gem-1-bulk_predict-metadata`**: Bulk RNA-seq metadata prediction model
- **`gem-1-sc_predict-metadata`**: Single-cell RNA-seq metadata prediction model

See the [Metadata Prediction](metadata-prediction.html) vignette for detailed usage.

Only baseline models are available to all users. You can check which models are available programmatically, use `list_models()`. Contact us at support@synthesize.bio if you have any questions.

### Listing Available Models

You can check which models are available programmatically:

```{r models}
# Check available models
list_models()
```

### Exploring Available Metadata

Each model accepts a specific set of metadata fields with defined vocabularies (valid ontology IDs, cell lines, tissues, etc.). You can browse and download these vocabularies at [app.synthesize.bio/docs/vocab](https://app.synthesize.bio/docs/vocab).

See the [Available Metadata](available-metadata.html) vignette for more details.

## Quick Start

Here's a quick example using a baseline model:
```{r query, eval=FALSE}
# Get an example query structure
query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Submit the query and get results
result <- predict_query(query, model_id = "gem-1-bulk")

# Access the results
metadata <- result$metadata
expression <- result$expression
```

For more detailed examples and advanced usage, see the model-specific vignettes linked above.

## Session info

```{r session-info}
sessionInfo()
```

## Additional Resources

- [Package Source Code](https://github.com/synthesizebio/rsynthbio)
- [File Bug Reports](https://github.com/synthesizebio/rsynthbio/issues)

---
title: "Metadata Prediction"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Metadata Prediction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE # Set to FALSE since API calls require credentials
)
```

## Overview

Metadata prediction models **infer biological metadata from observed expression data**. Given a gene expression profile, the model predicts the likely biological characteristics such as cell type, tissue, disease state, and more.

This is useful when you want to:

- Annotate samples of unknown origin
- Validate sample labels against expression patterns
- Discover potential mislabeled or contaminated samples
- Understand the biological characteristics captured in expression data

## Available Models

- **`gem-1-bulk_predict-metadata`**: Bulk RNA-seq metadata prediction model
- **`gem-1-sc_predict-metadata`**: Single-cell RNA-seq metadata prediction model

> **Note:** These endpoints may require 1-2 minutes of startup time if they have been scaled down. Plan accordingly for interactive use.

```{r}
library(rsynthbio)
```

## How It Works

Metadata prediction encodes your expression data into the model's latent space and then uses classifiers to predict the most likely metadata values for each sample. The model returns:

1. **Classifier probabilities**: For each categorical metadata field, the probability distribution over possible values
2. **Predicted labels**: The most likely value for each metadata field
3. **Latent representations**: The biological, technical, and perturbation latent vectors

## Creating a Query

Metadata prediction queries are simpler than other model types—you only need to provide expression counts:

```{r query-example, eval=FALSE}
# Get the example query structure
example_query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query

# Inspect the query structure
str(example_query)
```

The query structure includes:

1. **`inputs`**: A list of count vectors, where each element is a named list with a `counts` field containing expression values

2. **`seed`** (optional): Random seed for reproducibility

## Example: Predicting Sample Metadata

Here's a complete example predicting metadata for expression samples:

```{r predict-metadata, eval=FALSE}
# Start with example query structure
query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query

# Replace with your actual expression counts
# Each input should be a list with a counts vector
query$inputs <- list(
  list(counts = sample1_counts),
  list(counts = sample2_counts),
  list(counts = sample3_counts)
)

# Optional: set seed for reproducibility
query$seed <- 42

# Submit the query
result <- predict_query(query, model_id = "gem-1-bulk_predict-metadata")
```

## Example: Single Sample Prediction

For predicting metadata of a single sample:

```{r single-sample, eval=FALSE}
query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query

# Single sample
query$inputs <- list(
  list(counts = my_sample_counts)
)

result <- predict_query(query, model_id = "gem-1-bulk_predict-metadata")

# Access the predictions
print(result$outputs$metadata)
```

## Query Parameters

### inputs (list, required)

A list of expression count vectors. Each element should be a named list containing:

- **`counts`**: A vector of non-negative integers representing gene expression counts

```{r inputs-example, eval=FALSE}
query$inputs <- list(
  list(counts = c(0, 12, 5, 0, 33, 7, ...)), # Sample 1
  list(counts = c(3, 0, 0, 7, 1, 0, ...)) # Sample 2
)
```

### seed (integer, optional)

Random seed for reproducibility.

```{r seed-example, eval=FALSE}
query$seed <- 123
```

## Understanding the Results

The results from metadata prediction include several components:

### Predicted Metadata

The `metadata` data frame contains the predicted values for each sample:

```{r metadata-results, eval=FALSE}
# View predicted metadata
head(result$outputs$metadata)

# Access specific predictions
result$outputs$metadata$cell_type_ontology_id
result$outputs$metadata$tissue_ontology_id
result$outputs$metadata$disease_ontology_id
```

### Classifier Probabilities

For categorical metadata fields, the model returns probability distributions over all possible values. These are useful for understanding prediction confidence:

```{r classifier-probs, eval=FALSE}
# If probabilities are included in the output
# Access cell type probabilities for first sample
# The exact structure depends on the API response format

# Example: viewing top predicted cell types
cell_type_probs <- result$outputs$classifier_probs$cell_type[[1]]
head(sort(cell_type_probs, decreasing = TRUE))
```

### Latent Representations

The model also returns latent vectors that capture biological, technical, and perturbation characteristics:

```{r latents, eval=FALSE}
# Access latent representations (if returned)
biological_latents <- result$outputs$latents$biological
technical_latents <- result$outputs$latents$technical
```

## Use Cases

### Sample Annotation

Annotate unlabeled samples with predicted metadata:

```{r annotation, eval=FALSE}
# Load your unlabeled samples
unlabeled_counts <- read.csv("unlabeled_samples.csv", row.names = 1)

# Create query
query <- get_example_query(model_id = "gem-1-bulk_predict-metadata")$example_query
query$inputs <- lapply(1:ncol(unlabeled_counts), function(i) {
  list(counts = unlabeled_counts[, i])
})

# Predict metadata
result <- predict_query(query, model_id = "gem-1-bulk_predict-metadata")

# Combine with sample IDs
annotations <- result$outputs$metadata
annotations$sample_id <- colnames(unlabeled_counts)
```

### Quality Control

Validate existing sample labels against predicted metadata:

```{r qc, eval=FALSE}
# Compare predicted vs. provided labels
provided_labels <- c("UBERON:0002107", "UBERON:0002107", "UBERON:0000955", "UBERON:0000955")
predicted_labels <- result$outputs$metadata$tissue_ontology_id

# Identify potential mismatches
mismatches <- which(provided_labels != predicted_labels)
if (length(mismatches) > 0) {
  message("Potential mislabeled samples: ", paste(mismatches, collapse = ", "))
}
```

## Important Notes

### Counts Vector Length

The counts vector for each sample must match the model's expected number of genes. If the length doesn't match, the API will return a validation error.

Use `get_example_query()` to see the expected structure.

### Gene Order

Ensure your counts are in the same gene order expected by the model. The gene order should match what the baseline model expects—you can retrieve this from any prediction result's `gene_order` field.

### Non-Negative Counts

All count values must be non-negative integers. Floats that are whole numbers (like `10.0`) are accepted, but negative values will cause validation errors.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{data}
\name{DEFAULT_TIMEOUT}
\alias{DEFAULT_TIMEOUT}
\title{Default Timeout}
\format{
An object of class \code{numeric} of length 1.
}
\usage{
DEFAULT_TIMEOUT
}
\description{
Default timeout (seconds) for outbound HTTP requests
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key-handlers.R
\name{load_synthesize_token_from_keyring}
\alias{load_synthesize_token_from_keyring}
\title{Load Synthesize Bio API Token from Keyring}
\usage{
load_synthesize_token_from_keyring()
}
\value{
Invisibly returns TRUE if successful, FALSE if token not found in keyring.
}
\description{
Loads the previously stored Synthesize Bio API token from the system
keyring and sets it in the environment for the current session.
}
\examples{
\dontrun{
# Load token from keyring
load_synthesize_token_from_keyring()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key-handlers.R
\name{clear_synthesize_token}
\alias{clear_synthesize_token}
\title{Clear Synthesize Bio API Token}
\usage{
clear_synthesize_token(remove_from_keyring = FALSE)
}
\arguments{
\item{remove_from_keyring}{Logical, whether to also remove the token from the
system keyring if it's stored there. Defaults to FALSE.}
}
\value{
Invisibly returns TRUE.
}
\description{
Clears the Synthesize Bio API token from the environment for the
current R session. This is useful for security purposes when you've finished
working with the API or when switching between different accounts.
}
\examples{
\dontrun{
# Clear token from current session only
clear_synthesize_token()

# Clear token from both session and keyring
clear_synthesize_token(remove_from_keyring = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key-handlers.R
\name{set_synthesize_token}
\alias{set_synthesize_token}
\title{Set Synthesize Bio API Token}
\usage{
set_synthesize_token(use_keyring = FALSE, token = NULL)
}
\arguments{
\item{use_keyring}{Logical, whether to also store the token securely in the system
keyring for future sessions. Defaults to FALSE.}

\item{token}{Character, optional. If provided, uses this token instead of prompting.
This parameter should only be used in non-interactive scripts.}
}
\value{
Invisibly returns TRUE if successful.
}
\description{
Securely prompts for and stores the Synthesize Bio API token in the
environment. This function uses getPass to securely handle the token input without
displaying it in the console. The token is stored in the SYNTHESIZE_API_KEY
environment variable for the current R session.
}
\examples{
# Interactive prompt for token
\dontrun{
set_synthesize_token()

# Provide token directly (less secure, not recommended for interactive use)
set_synthesize_token(token = "your-token-here")

# Store in system keyring for future sessions
set_synthesize_token(use_keyring = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/call_model_api.R
\name{get_example_query}
\alias{get_example_query}
\title{Get Example Query for Model}
\usage{
get_example_query(model_id, api_base_url = NULL, self_hosted = NULL)
}
\arguments{
\item{model_id}{Character string specifying the model ID (e.g., "gem-1-bulk", "gem-1-sc").}

\item{api_base_url}{The base URL for the API server. When NULL (default), it
is resolved from the `SYNTHESIZE_API_BASE_URL` environment variable,
falling back to the production default (API_BASE_URL). Point this at a
self-hosted model container to fetch its example query.}

\item{self_hosted}{Logical; when TRUE, the request targets a self-hosted
container and does not require an API key (one is only sent if set).
When NULL (default), it is resolved from the `SYNTHESIZE_SELF_HOSTED`
environment variable (truthy for 1/true/yes/on), defaulting to FALSE.}
}
\value{
A list representing a valid query structure for the specified model.
}
\description{
Retrieves an example query structure for a specific model.
This provides a template that can be modified for your specific needs.
}
\examples{
\dontrun{
# Get example query for bulk RNA-seq model
query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Get example query for single-cell model
query_sc <- get_example_query(model_id = "gem-1-sc")$example_query

# Modify the query structure
query$inputs[[1]]$num_samples <- 10

# Fetch from a self-hosted container (no API key required)
query <- get_example_query(
  model_id = "gem-1-bulk",
  api_base_url = "https://gem-1-bulk.internal.partner.example",
  self_hosted = TRUE
)$example_query
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/call_model_api.R
\name{list_models}
\alias{list_models}
\title{List Available Models}
\usage{
list_models(api_base_url = NULL, self_hosted = NULL)
}
\arguments{
\item{api_base_url}{The base URL for the API server. When NULL (default), it
is resolved from the `SYNTHESIZE_API_BASE_URL` environment variable,
falling back to the production default (API_BASE_URL). Point this at a
self-hosted model container to list its models.}

\item{self_hosted}{Logical; when TRUE, the request targets a self-hosted
container and does not require an API key (one is only sent if set).
When NULL (default), it is resolved from the `SYNTHESIZE_SELF_HOSTED`
environment variable (truthy for 1/true/yes/on), defaulting to FALSE.}
}
\value{
A list or data frame containing available models with their IDs and metadata.
}
\description{
Returns a list of all models available in the Synthesize Bio API.
Each model has a unique ID that can be used with predict_query() and get_example_query().
}
\examples{
\dontrun{
# Get all available models
models <- list_models()
print(models)

# List models from a self-hosted container (no API key required)
models <- list_models(
  api_base_url = "https://gem-1-bulk.internal.partner.example",
  self_hosted = TRUE
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{data}
\name{API_BASE_URL}
\alias{API_BASE_URL}
\title{API Base URL}
\format{
An object of class \code{character} of length 1.
}
\usage{
API_BASE_URL
}
\description{
Base URL for the Synthesize Bio API
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/call_model_api.R
\name{predict_query}
\alias{predict_query}
\title{Predict Gene Expression}
\usage{
predict_query(
  query,
  model_id,
  api_base_url = NULL,
  poll_interval_seconds = DEFAULT_POLL_INTERVAL_SECONDS,
  poll_timeout_seconds = DEFAULT_POLL_TIMEOUT_SECONDS,
  return_download_url = FALSE,
  raw_response = FALSE,
  self_hosted = NULL,
  ...
)
}
\arguments{
\item{query}{A list representing the query data to send to the API.
Use `get_example_query()` to generate an example. The query supports additional
optional fields:
\itemize{
  \item `total_count` (integer): Library size used when converting predicted log CPM
        back to raw counts. Higher values scale counts up proportionally.
  \item `deterministic_latents` (logical): If TRUE, the model uses the mean of each
        latent distribution instead of sampling, producing deterministic outputs for
        the same inputs. Useful for reproducibility.
  \item `seed` (integer): Random seed for reproducibility.
}}

\item{model_id}{Character string specifying the model ID (e.g., "gem-1-bulk", "gem-1-sc").
Use `list_models()` to see available models.}

\item{api_base_url}{The base URL for the API server. When NULL (default), it
is resolved in order from the per-model environment variable
`SYNTHESIZE_API_BASE_URL__<MODEL>` (e.g.
`SYNTHESIZE_API_BASE_URL__GEM_1_BULK`), then the global
`SYNTHESIZE_API_BASE_URL`, then the production default (API_BASE_URL).
The per-model variable lets you point each self-hosted model at its own
container once and omit `api_base_url` on every call.}

\item{poll_interval_seconds}{Seconds between polling attempts of the status endpoint.
Default is DEFAULT_POLL_INTERVAL_SECONDS (2).}

\item{poll_timeout_seconds}{Maximum total seconds to wait before timing out.
Default is DEFAULT_POLL_TIMEOUT_SECONDS (900 = 15 minutes).}

\item{return_download_url}{Logical, if TRUE, returns a list containing the signed
download URL instead of parsing into data frames. Default is FALSE.}

\item{raw_response}{Logical, if TRUE, returns the raw (unformatted) response
from the API without applying any output transformers. For the
production path this is the parsed JSON; for `self_hosted = TRUE` it is
the parsed Arrow `Table` together with its schema metadata. Default is FALSE.}

\item{self_hosted}{Logical, if TRUE, sends a single synchronous request to a
self-hosted model container that returns predictions as an Apache Arrow
IPC stream (no polling, no download URL). Requires the optional `arrow`
package and an `api_base_url` pointing at the container. Unlike the
production path, no API key is required (one is only sent if configured).
When NULL (default), it is resolved from the `SYNTHESIZE_SELF_HOSTED`
environment variable (truthy for 1/true/yes/on), defaulting to FALSE.}

\item{...}{Additional parameters to include in the query body. These are passed
directly to the API and validated server-side.}
}
\value{
A list. For the production path, if `return_download_url` is `FALSE`
        (default) the list contains `metadata` and `expression` data frames;
        if `TRUE` it contains `download_url` and empty data frames. For
        `self_hosted = TRUE`, the list contains the transformed data frames
        (`metadata`, `expression`, and `latents`; plus `classifier_probs` for
        metadata-prediction models) with `model_version` and `request_type`
        attached as attributes.
}
\description{
Sends a query to the Synthesize Bio API for prediction
and retrieves gene expression samples. This function sends the query
to the API and processes the response into usable data frames.
}
\examples{
# Set your API key (in practice, use a more secure method)
\dontrun{

# To start using rsynthbio, first you need to have an account with synthesize.bio.
# Go here to create one: https://app.synthesize.bio/

set_synthesize_token()

# Get available models
models <- list_models()

# Create a query for a specific model
query <- get_example_query(model_id = "gem-1-bulk")$example_query

# Request raw counts
result <- predict_query(query, model_id = "gem-1-bulk")

# Access the results
metadata <- result$metadata
expression <- result$expression

# Explore the top expressed genes in the first sample
head(sort(expression[1, ], decreasing = TRUE))

# Use deterministic latents for reproducible results
query$deterministic_latents <- TRUE
result_det <- predict_query(query, model_id = "gem-1-bulk")

# Specify a custom total count (library size)
query$total_count <- 5000000
result_custom <- predict_query(query, model_id = "gem-1-bulk")

# Self-hosted container returning a synchronous Apache Arrow IPC stream
result_sh <- predict_query(
  query,
  model_id = "gem-1-bulk",
  api_base_url = "https://gem-1-bulk.internal.partner.example",
  self_hosted = TRUE
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{data}
\name{SELF_HOSTED_TIMEOUT}
\alias{SELF_HOSTED_TIMEOUT}
\title{Self-Hosted Timeout}
\format{
An object of class \code{numeric} of length 1.
}
\usage{
SELF_HOSTED_TIMEOUT
}
\description{
Timeout (seconds) for synchronous self-hosted container
predictions. These run on the partner's GPU box and can take minutes for
large sample counts, so they use a longer timeout than hosted control-plane
calls.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key-handlers.R
\name{has_synthesize_token}
\alias{has_synthesize_token}
\title{Check if Synthesize Bio API Token is Set}
\usage{
has_synthesize_token()
}
\value{
Logical, TRUE if token is set, FALSE otherwise.
}
\description{
Checks whether a Synthesize Bio API token is currently set in the
environment. Useful for conditional code that requires an API token.
}
\examples{
 \dontrun{
# Check if token is set
if (!has_synthesize_token()) {
  # Prompt for token if not set
  set_synthesize_token()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{data}
\name{DEFAULT_POLL_INTERVAL_SECONDS}
\alias{DEFAULT_POLL_INTERVAL_SECONDS}
\title{Default Poll Interval}
\format{
An object of class \code{numeric} of length 1.
}
\usage{
DEFAULT_POLL_INTERVAL_SECONDS
}
\description{
Default polling interval (seconds) for async model queries
}
\keyword{datasets}
